from casm.bset.cluster_functions import BasisFunctionSpecs


def test_basis_function_specs_1():
    basis_function_specs = BasisFunctionSpecs(
        dofs=["occ", "Hstrain"],
        dof_specs={
            "occ": {
                "site_basis_functions": "occupation",
            }
        },
        orbit_branch_max_poly_order={
            # use maximum polynomial order 6,
            # for orbits of cluster size 1
            "1": 6,
            # use maximum polynomial order 5,
            # for orbits of cluster size 4
            "2": 5,
        },
        # use max(3, cluster size) for all other orbits
        global_max_poly_order=3,
    )

    assert isinstance(basis_function_specs, BasisFunctionSpecs)

    data = basis_function_specs.to_dict()
    assert isinstance(data, dict)
    assert data["dofs"] == ["occ", "Hstrain"]
    assert data["orbit_branch_max_poly_order"] == {"1": 6, "2": 5}
    assert data["global_max_poly_order"] == 3
    assert data["dof_specs"] == {"occ": {"site_basis_functions": "occupation"}}
    assert "default" not in data

    basis_function_specs_in = BasisFunctionSpecs.from_dict(data)
    assert isinstance(basis_function_specs_in, BasisFunctionSpecs)
    assert basis_function_specs_in.dofs == ["occ", "Hstrain"]
    assert basis_function_specs_in.dof_specs == {
        "occ": {"site_basis_functions": "occupation"}
    }
    assert basis_function_specs_in.global_max_poly_order == 3
    assert basis_function_specs_in.orbit_branch_max_poly_order == {"1": 6, "2": 5}
    assert basis_function_specs_in.param_pack_type == "default"


def test_basis_function_specs_2():
    basis_function_specs = BasisFunctionSpecs()

    assert isinstance(basis_function_specs, BasisFunctionSpecs)

    data = basis_function_specs.to_dict()
    assert isinstance(data, dict)
    assert "dofs" not in data
    assert "orbit_branch_max_poly_order" not in data
    assert "global_max_poly_order" not in data
    assert "dof_specs" not in data
    assert "param_pack_type" not in data

    basis_function_specs_in = BasisFunctionSpecs.from_dict(data)
    assert isinstance(basis_function_specs_in, BasisFunctionSpecs)
    assert basis_function_specs_in.dofs is None
    assert basis_function_specs_in.dof_specs == {}
    assert basis_function_specs_in.global_max_poly_order is None
    assert basis_function_specs_in.orbit_branch_max_poly_order == {}
    assert basis_function_specs_in.param_pack_type == "default"



import copy
import sys
import time
from typing import Callable, Optional

import libcasm.clusterography as casmclust
import libcasm.configuration as casmconfig
from casm.bset.clexwriter._cpp_str import (
    CppFormatProperties,
)
from casm.bset.cluster_functions import get_occ_site_functions
from casm.bset.polynomial_functions import (
    PolynomialFunction,
)
from libcasm.clexulator import (
    PrimNeighborList,
)

from ._methods import (
    make_neighborhoods,
    make_orbit_bfuncs,
    make_site_bfuncs,
)


class WriterV1Basic:
    """Generates the variables needed for the `v1.basic` Clexulator, the
    standard CASM v1 compatible Clexulator (no automatic differentiation)"""

    @staticmethod
    def template_name():
        """The name of the template file used by this writer

        Returns
        -------
        template_name: str

            The name of the template file used by this writer, as a path relative to
            the casm.bset template directory.
        """
        return "v1.basic/Clexulator.cpp"

    def __init__(
        self,
        prim: casmconfig.Prim,
        project_name: str,
        bset_name: str,
        is_periodic: bool,
        i_clex: Optional[int],
        prim_neighbor_list: PrimNeighborList,
        clusters: list[list[casmclust.Cluster]],
        functions: list[list[list[PolynomialFunction]]],
        occ_site_functions: list[dict],
        occ_site_functions_info: dict,
        linear_function_indices: Optional[set[int]] = None,
        make_variable_name_f: Optional[Callable] = None,
        local_discrete_dof: Optional[dict] = None,
        cpp_fmt: Optional[CppFormatProperties] = None,
        verbose: bool = False,
    ):
        """

        .. rubric:: Constructor

        Parameters
        ----------
        prim: libcasm.configuration.Prim
            The prim.
        project_name: str
            Project name. Used to construct the Clexulator class name. This must
            consist of alphanumeric characters and underscores only. The first
            character may not be a number.
        bset_name: str
            Basis set name. Used to construct the Clexulator class name. This must
            consist of alphanumeric characters and underscores only.
        is_periodic: bool
            Whether the Clexulator is for a periodic basis set or local basis set.
        i_clex: Optional[int]
            If a local basis set, the equivalent clexulator index (start from 0); or
            None to write the prototype.
        prim_neighbor_list: libcasm.clexulator.PrimNeighborList
            The :class:`~libcasm.clexulator.PrimNeighborList` which has been used to
            set the `neighborhood_site_index` for the
            :class:`~casm.bset.polynomial_functions.Variable` used by the `functions`.
        clusters: list[list[libcasm.clusterography.Cluster]]
            The clusters for which cluster functions have been constructed. The cluster
            ``clusters[i_orbit][i_equiv]`` is the `i_equiv`-th symmetrically
            equivalent cluster in the `i_orbit`-th orbit. The order of sites in the
            clusters is not arbitrary, it must be consistent with the
            `cluster_site_index` of the
            :class:`~casm.bset.polynomial_functions.Variable` used in the
            :class:`~casm.bset.polynomial_functions.PolynomialFunction` stored in
            :py:data:`~ClusterFunctionsBuilder.functions`. This can be generated by
            :class:`~casm.bset.cluster_functions.ClusterFunctionsBuilder`.
        functions: list[list[list[casm.bset.polynomial_functions.PolynomialFunction]]]
            List of functions, as generated by
            :class:`~casm.bset.ClusterFunctionsBuilder`, where function
            ``functions[i_orbit][i_cluster][i_func]`` is the `i_func`-th function
            on the `i_cluster`-th cluster in the `i_orbit`-th orbit.
        occ_site_functions: list[dict]
            List of occupation site basis functions. For each sublattice with discrete
            site basis functions, must include:

            - `"sublattice_index"`: int, index of the sublattice
            - `"value"`: list[list[float]], list of the site basis function values,
              as ``value[function_index][occupant_index]``.

        occ_site_functions_info: dict
            Information about occupation site basis functions.

            Occupation site basis functions info, with format:

            - `"max_function_index"`: int, The maximum site function index, across all
              sublattices.
            - `"all_sublattices_have_same_site_functions"`: bool, True if all
              sublattices have same site functions; False otherwise.
            - `"occ_var_name"`: str, A variable name template for the site functions,
              which may be formated using `b` for sublattice index and `m` for site
              function index (i..e ``occ_var_name.format(b=0, m=1)``).
            - `"occ_var_desc"`: str, A description of the occupation
              variable, including a description of the subscript indices.
            - `"occ_var_indices"`: list[list[str, str]], A list of lists, where each
              sublist contains the variable name and description for each subscript
              index.

        linear_function_indices : Optional[set[int]]
            The linear indices of the functions that will be included. If None,
            all functions will be included in the Clexulator. Otherwise,
            only the specified functions will be included in the Clexulator. This
            is an experimental feature.

        make_variable_name_f: Optional[Callable] = None
            Allows specifying a custom class to construct variable names. The default
            class used is :class:`~casm.bset.cluster_functions.MakeVariableName`.
            Custom classes should have the same `__call__` signature as
            :class:`~casm.bset.cluster_functions.MakeVariableName`, and have
            `occ_var_name` and `occ_var_desc` attributes.

        local_discrete_dof: Optional[list[str]] = None
            The types of local discrete degree of freedom (DoF).

        cpp_fmt: Optional[CppFormatProperties] = None,
            C++ string formatting properties. If None, the default values are used.

            .. code-block:: Python

                CppFormatProperties(
                    coeff_fmt_spec=".10f",
                    coeff_atol=1e-10,
                )

        verbose: bool = False
            Print progress statements

        """

        ### defaults ###
        if cpp_fmt is None:
            cpp_fmt = CppFormatProperties()

        ### set variables ###

        # set clexulator_name
        self.clexulator_name = None
        """str: Name of the Clexulator.
        
        This is used for the Clexulator class and must be a valid C++ identifier. It
        is also used for the Clexulator source file name, with extension ".cpp" added.
        The normal convention is that it is:
        
        - For periodic / local prototype: "{project_name}_Clexulator_{bset_name}"
        - For local equivalents: "{project_name}_Clexulator_{bset_name}_{i_clex}"
        
        """

        if is_periodic or i_clex is None:
            self.clexulator_name = f"{project_name}_Clexulator_{bset_name}"
        else:
            self.clexulator_name = f"{project_name}_Clexulator_{bset_name}_{i_clex}"

        self.prim_neighbor_list = prim_neighbor_list
        """libcasm.clexulator.PrimNeighborList: The PrimNeighborList which has been \
        used to set the `neighborhood_site_index` for the 
        :class:`~casm.bset.polynomial_functions.Variable` used by the `functions`."""

        self.occ_site_functions = copy.deepcopy(occ_site_functions)
        """list[dict]: List of occupation site basis functions. 
        
        For each sublattice with discrete site basis functions, includes:

        - `"sublattice_index"`: int, index of the sublattice
        - `"n_occupants"`: int, number of allowed occupants
        - `"value"`: list[list[float]], list of the site basis function values,
          as ``value[function_index][occupant_index]``
        - `"constant_function_index"`: int, index of the constant site basis function
          
        """
        for site_funcs in self.occ_site_functions:
            site_funcs["n_occupants"] = len(site_funcs["value"][0])

            for i in range(len(site_funcs["value"])):
                # note constant occupation functions
                phi = get_occ_site_functions(
                    occ_site_functions=occ_site_functions,
                    sublattice_index=site_funcs["sublattice_index"],
                    site_function_index=i,
                )

                if (phi == 1.0).all():
                    site_funcs["constant_function_index"] = i

            if "constant_function_index" not in site_funcs:
                raise ValueError(
                    f"Occupation site basis functions for sublattice "
                    f"{site_funcs['sublattice_index']} must include "
                    "a constant function."
                )

        self.occ_site_functions_info = occ_site_functions_info
        """dict: Information about occupation site basis functions.
        
        Occupation site basis functions info, with format:
        
        - `"max_function_index"`: int, The maximum site function index, across all
            sublattices.
        - `"all_sublattices_have_same_site_functions"`: bool, True if all _sublattices
            have same site functions; False otherwise.
        - `"occ_var_name"`: str, A variable name template for the site functions,
            which may be formated using `b` for sublattice index and `m` for site 
            function index (i..e ``occ_var_name.format(b=0, m=1)``).
        - `"occ_var_desc"`: str, A description of the occupation
            variable, including a description of the subscript indices.
        - `"occ_var_indices"`: list[list[str, str]], A list of lists, where each sublist
          contains the variable name and description for each subscript index.
        
        """

        ## set linear_function_indices
        self.linear_function_indices = linear_function_indices
        """Optional[set[int]]: The linear indices of the functions that will be 
        included.

        If None, all functions will be included in the Clexulator. Otherwise, only the 
        specified functions will be included in the Clexulator. This
        is an experimental feature."""

        ## set cpp_fmt
        self.cpp_fmt = cpp_fmt
        """libcasm.bset.clexwriter.CppFormatProperties: C++ string formatting \
        properties."""

        ## set n_corr
        self.n_corr = None
        """int: Total number of correlations.

        This is the total number of correlations that could be calculated and does not
        change if `linear_function_indices` is used to prevent evaluation of 
        unnecessary correlations. 
        """

        self.n_corr = 1  # include constant term
        for i_orbit, orbit_functions in enumerate(functions):
            prototype_functions = orbit_functions[0]
            self.n_corr += len(prototype_functions)

        ## set neighborhood info
        if verbose:
            start = time.time()
            print("Generating neighborhoods...")
            sys.stdout.flush()

        _complete_neighborhood, _function_neighborhoods = make_neighborhoods(
            prim=prim,
            is_periodic=is_periodic,
            prim_neighbor_list=prim_neighbor_list,
            clusters=clusters,
            functions=functions,
            linear_function_indices=linear_function_indices,
            make_variable_name_f=make_variable_name_f,
            local_discrete_dof=local_discrete_dof,
        )

        if verbose:
            print("Generating neighborhoods DONE")
            elapsed_time = time.time() - start
            print(f"time: {elapsed_time:0.4f} (s)")
            print()
            sys.stdout.flush()

        self.complete_neighborhood = _complete_neighborhood
        """dict: The neighborhood needed to evaluate basis functions 
        involving sites in the origin unit cell. 
        
        Contains:
        
        - ``"unitcells"``: list[list[int]], List of ``[i,j,k]`` unit
          cell indices for the neighborhood.
        - ``"sites"``: list[list[int]], List of ``[b,i,j,k]`` site
          indices for the neighborhood.
        
        """

        self.function_neighborhoods = _function_neighborhoods
        """list[dict]: The neighborhood data for each orbit of basis functions. 
        
        Each dictionary contains:

        - ``"linear_function_index"``: int, Linear function index
        - ``"same_as"``: Optional[int], If not None, the linear function index of the
          first orbit of functions in the same cluster orbit as this function, which
          has the same neighborhood.
        - ``"unitcells"``: list[list[int]], List of ``[i,j,k]`` unit
          cell indices for the neighborhood of the orbit of basis function. This is
          only populated for the first function in each cluster orbit.
        - ``"sites"``: list[list[int]], List of ``[b,i,j,k]`` site
          indices for the neighborhood of the orbit of basis function. This is
          only populated for the first function in each cluster orbit.
        
        """

        ## set nlist_size
        import libcasm.xtal as xtal

        nlist_size = 0
        for bijk in self.complete_neighborhood["sites"]:
            site = xtal.IntegralSiteCoordinate.from_list(bijk)
            nlist_size = max(nlist_size, prim_neighbor_list.neighbor_index(site) + 1)
        self.nlist_size = nlist_size
        """int: Maximum neighbor list index of sites involved in evaluating basis 
        functions."""

        ## set orbit_bfuncs and orbit_bfuncs_variables_needed
        self.orbit_bfuncs = None
        """list[dict]: Data for each orbit basis function (i.e. if periodic, the 
        contribution to the global correlations from one unit cell for periodic; if
        local, the sum of all symmetrically equivalent local cluster functions). 
        
        Includes:

        - ``"linear_function_index"``: int, Linear function index
        - ``"linear_orbit_index"``: int, Linear cluster orbit index
        - ``"cpp"``: str, C++ expression for evaluating the global correlation
          contribution
        - ``"latex_prototype"``: str, Latex formula for the prototype cluster
        - ``"latex_orbit"``: str, Latex formula for the orbit contribution
        
        """

        self.orbit_bfuncs_variables_needed = None
        """dict[str,list[list[int]]]: The variables needed to evaluate the orbit basis 
        functions. 
        
        The dictionary keys are DoF type key, and values are lists of
        [`component_index`, `neighbor_list_index`, `sublattice_index`] for variables
        of that type. The `neighbor_list_index` and `sublattice_index` for global
        continuous DoF are None.
        """

        if verbose:
            start = time.time()
            print("Generating orbit function formulas...")
            sys.stdout.flush()

        self.orbit_bfuncs, self.orbit_bfuncs_variables_needed = make_orbit_bfuncs(
            prim_neighbor_list=self.prim_neighbor_list,
            clusters=clusters,
            functions=functions,
            occ_site_functions=self.occ_site_functions,
            cpp_fmt=self.cpp_fmt,
            linear_function_indices=self.linear_function_indices,
        )

        if verbose:
            print("Generating orbit function formulas DONE")
            elapsed_time = time.time() - start
            print(f"time: {elapsed_time:0.4f} (s)")
            print()
            sys.stdout.flush()

        ## set site_bfuncs and site_bfuncs_variables_needed_at

        self.site_bfuncs = None
        """list[dict]: List of dictionaries, one per cluster expansion basis function, 
        with information needed for printing the Clexulator point correlation 
        evaluation methods.
        
        The format is:

        - "linear_function_index": int - The cluster expansion basis function index.
        - "linear_orbit_index": int - The cluster orbit index
        - "at": list[dict] - Data for functions at specific points

          - "neighbor_list_index": int - Index into the neighbor list of the point
          - "cpp": str - The C++ code to evaluate the point functions
          - "occ_delta_cpp": str - The C++ code to evaluate the change in point
            functions due to the change of occupation index on the site from
            ``"occ_i"`` to ``"occ_f"``.
          - "latex_prototype": str - The latex expression for the point functions,
            including only functions from the prototype cluster
          - "latex_orbit": str - The latex expression for the point functions,
            including functions from the entire cluster orbit
        
        """

        self.site_bfuncs_variables_needed_at = None
        """list[dict[str,list[list[int]]]]: For each point, the variables needed to 
        evaluate the point functions. 
        
        The value of `site_bfuncs_variables_needed_at[neighbor_list_index]` is a 
        dictionary, where the keys are DoF type key, and values are lists of
        [`component_index`, `neighbor_list_index`, `sublattice_index`] for variables
        of that type. The `neighbor_list_index` and `sublattice_index` for global
        continuous DoF are None.
        """

        if verbose:
            start = time.time()
            print("Generating site function formulas...")
            sys.stdout.flush()

        self.site_bfuncs, self.site_bfuncs_variables_needed_at = make_site_bfuncs(
            prim=prim,
            is_periodic=is_periodic,
            prim_neighbor_list=self.prim_neighbor_list,
            clusters=clusters,
            functions=functions,
            occ_site_functions=self.occ_site_functions,
            cpp_fmt=self.cpp_fmt,
            linear_function_indices=self.linear_function_indices,
            make_variable_name_f=make_variable_name_f,
            local_discrete_dof=local_discrete_dof,
        )

        if verbose:
            print("Generating site function formulas DONE")
            elapsed_time = time.time() - start
            print(f"time: {elapsed_time:0.4f} (s)")
            print()
            sys.stdout.flush()

        ## set n_point_corr_sites
        n_point_corr_sites = 0
        for x in self.site_bfuncs:
            for y in x["at"]:
                if y["n_point_functions"] > 0:
                    n_point_corr_sites = max(
                        n_point_corr_sites, y["neighbor_list_index"] + 1
                    )
        self.n_point_corr_sites = n_point_corr_sites
        """int: Number of sites where point correlations can be evaluated.

        This is the maximum neighbor list index for sites that have site functions.
        For periodic cluster expansions, the maximum value this can take is the number 
        of sublattices included in the `prim_neighbor_list`. For local cluster 
        expansions, the maximum value this can take is the maximum neighbor list index
        of the complete site neighborhood. It is possible that some of the sites do not 
        actually have any site functions associated with them, in which case the point 
        correlations all evaluate as zero.
        """

        # if is_periodic:
        #     self.n_point_corr_sites = len(prim_neighbor_list.sublattice_indices())
        # else:
        #     self.n_point_corr_sites = prim_neighbor_list.n_neighborhood_sites()

        ## set continuous DoF and params (parameter pack implementation details)

        self.continuous_dof = []
        """list[dict]: List of continuous DoF to register.

        For each, must include:

        - `"key"`: str, DoF name (i.e. ``"Hstrain``)
        - `"is_global"`: bool, DoF type
        - `"max_n_components"`: int, Maximum number of components (for local DoF)
        - `"sites"`: list[dict], Describes local DoF, includes:

          - `"sublattice_index"`: int, Sublattice index containing local DoF
          - `"n_components"`: int, Number of components on this sublattice

        """

        self.params = []
        """list[dict]: List of parameter packs to include.

        Parameter packs are a data structure used internally by the Clexulator that
        store the evaluated variables used by the cluster functions. The data structure
        type can be written in a generic fashion to allow automatic differentiation
        when it is enabled. There should be one parameter pack for each DoF included 
        in the basis functions and one for correlations. For each parameter pack, must 
        include:

        - `"name"`: str, Parameter name (i.e. ``"occ_site_func"``,
          ``"Hstrain_var"``, ``"disp_var"``, or ``"corr"``.)
        - `"rows"`: int, Number of rows needed by the parameter pack. For DoF,
          this is the maximum number of components. For "corr", this is the number
          of functions.
        - `"cols"`: int, Number of columns needed by the parameter pack. For site
          DoF this is the number of neighbors, for global DoF this is 1,
          and for "corr" this is 1.
        - `"is_independent"`: str, String representation of C++ bool. For DoF
          this is ``"true"``, and for "corr" this is ``"false"``.

        """

        # add correlations parameter pack
        corr_param = {
            "name": "corr",
            "rows": self.n_corr,
            "cols": 1,
            "is_independent": "false",
        }
        self.params.append(corr_param)

        # add dof parameter packs and continuous dof info
        for key, variables_needed in self.orbit_bfuncs_variables_needed.items():
            if len(variables_needed) == 0:
                continue

            # check variables of type==key
            is_global = True
            max_component_index = 0
            max_component_index_by_site = {}
            for var in variables_needed:
                component_index = var[0]
                sublattice_index = var[2]
                if sublattice_index is not None:
                    is_global = False
                if component_index > max_component_index:
                    max_component_index = component_index
                if sublattice_index is not None:
                    if sublattice_index not in max_component_index_by_site:
                        max_component_index_by_site[sublattice_index] = 0
                    if component_index > max_component_index_by_site[sublattice_index]:
                        max_component_index_by_site[sublattice_index] = component_index
            rows = max_component_index + 1

            if is_global:
                # global Dof only need one column in parameter pack
                cols = 1
                sites_info = None
            else:
                # local Dof need one column per neighborhood site in parameter pack
                cols = self.prim_neighbor_list.n_neighborhood_sites()

                # get max number of components by sublattice
                sites_info = []
                for b, c_max in max_component_index_by_site.items():
                    sites_info.append(
                        {
                            "sublattice_index": b,
                            "n_components": c_max + 1,
                        }
                    )

            if key == "occ":
                is_discrete = True
                name = "occ_site_func"
            else:
                is_discrete = False
                name = f"{key}_var"

            dof_param = {
                "name": name,
                "rows": rows,
                "cols": cols,
                "is_independent": "true",
            }
            self.params.append(dof_param)

            if not is_discrete:
                dof_info = {
                    "key": key,
                    "is_global": is_global,
                }
                if not is_global:
                    dof_info["max_n_components"] = rows
                    dof_info["sites"] = sites_info
                self.continuous_dof.append(dof_info)

        ## set cpp_fmt
        self.cpp_fmt = cpp_fmt
        """casm.bset.clexwriter.CppFormatProperties: C++ string formatting properties.
        """

    def variables(self) -> dict:
        """Returns a dictionary with variables to be rendered using a Jinja2 template.

        Returns
        -------
        variables: dict
            Variables used by the v1.basic/Clexulator.cpp Jinja2 Clexulator template

        """
        return {
            "clexulator_name": self.clexulator_name,
            "nlist_size": self.nlist_size,
            "n_corr": self.n_corr,
            "n_point_corr_sites": self.n_point_corr_sites,
            "orbit_bfuncs": self.orbit_bfuncs,
            "orbit_bfuncs_variables_needed": self.orbit_bfuncs_variables_needed,
            "site_bfuncs": self.site_bfuncs,
            "site_bfuncs_variables_needed_at": self.site_bfuncs_variables_needed_at,
            "occ_site_functions": self.occ_site_functions,
            "occ_site_functions_info": self.occ_site_functions_info,
            "continuous_dof": self.continuous_dof,
            "params": self.params,
            "nlist_weight_matrix": self.prim_neighbor_list.weight_matrix().tolist(),
            "nlist_sublattice_indices": self.prim_neighbor_list.sublattice_indices(),
            "nlist_total_n_sublattice": self.prim_neighbor_list.total_n_sublattice(),
            "complete_neighborhood": self.complete_neighborhood,
            "function_neighborhoods": self.function_neighborhoods,
        }



"""Simple top-level casm.bset methods for the most common use cases"""

import pathlib
from typing import Any, Optional, Union

import numpy as np

import casm.bset._helpers as _helpers
import libcasm.configuration as casmconfig
import libcasm.xtal as xtal
from casm.bset.clexwriter import (
    ClexulatorWriter,
    CppFormatProperties,
)
from casm.bset.cluster_functions import (
    BasisFunctionSpecs,
    ClexBasisSpecs,
    ClusterFunctionsBuilder,
)
from libcasm.clexulator import (
    Correlations,
    PrimNeighborList,
    SuperNeighborList,
    make_clexulator,
)
from libcasm.clusterography import (
    Cluster,
    ClusterOrbitGenerator,
    ClusterSpecs,
    make_cluster_group,
    make_integral_site_coordinate_symgroup_rep,
)
from libcasm.occ_events import (
    OccEvent,
    make_occevent_group,
    make_occevent_symgroup_rep,
)


def make_clex_basis_specs(
    prim: Union[xtal.Prim, casmconfig.Prim, dict, str, pathlib.Path],
    dofs: Optional[list[str]] = None,
    max_length: Optional[list[float]] = [],
    custom_generators: Optional[list[ClusterOrbitGenerator]] = [],
    phenomenal: Union[Cluster, OccEvent, None] = None,
    cutoff_radius: Optional[list[float]] = [],
    occ_site_basis_functions_specs: Any = None,
    global_max_poly_order: Optional[int] = None,
    orbit_branch_max_poly_order: Optional[dict] = None,
) -> ClexBasisSpecs:
    """Constructs cluster expansion basis functions specifications

    Parameters
    ----------
    prim: Union[libcasm.xtal.Prim, libcasm.configuration.Prim, dict, str, pathlib.Path]
        The prim, with symmetry information. May be provided as a Prim instance, a Prim
        dict, or the path to a file containing the Prim dict.

    dofs: Optional[list[str]] = None
        An list of string of dof type names that should be used to construct basis
        functions. The default value is all DoF types included in the prim.

    max_length: list[float] = []
        The maximum site-to-site distance to allow in clusters, by number of sites in
        the cluster. Example: `[0.0, 0.0, 5.0, 4.0]` specifies that pair clusters up to
        distance 5.0 and triplet clusters up to distance 4.0 should be included. The
        null cluster and point cluster values (elements 0 and 1) are arbitrary.

    custom_generators: list[libcasm.clusterography.ClusterOrbitGenerator] = []]
        Specifies clusters that should be uses to construct orbits regardless of the
        `max_length` or `cutoff_radius` parameters.

    phenomenal: Union[libcasm.clusterography.Cluster, libcasm.occ_events.OccEvent, \
    None] = None
        If provided, generate local cluster functions using the invariant group of the
        phenomenal cluster or event. By default, periodic cluster functions are
        generated.

    cutoff_radius: list[float] = []
        For local clusters, the maximum distance of sites from any phenomenal cluster
        site to include in the local environment, by number of sites in the cluster.
        The null cluster value (element 0) is arbitrary.

    occ_site_basis_functions_specs: Any = None
        Provides instructions for constructing occupation site basis functions.
        The most common options are "chebychev" or "occupation". This
        parameter corresponds to the value of

        .. code-block:: Python

            "dof_specs": {
                "occ": {
                    "site_basis_functions": ...
                }
            }

        as described in detail in the section
        :ref:`DoF Specifications <sec-dof-specifications>` and is required for
        functions of occupation DoF.

    global_max_poly_order: Optional[int] = None
        The maximum order of polynomials of continuous DoF to generate, for any
        orbit not specified more specifically by `orbit_branch_max_poly_order`.

    orbit_branch_max_poly_order: Optional[dict[int, int]] = None
        Specifies for continuous DoF the maximum polynomial order to generate by
        cluster size, according to
        ``orbit_branch_max_poly_order[cluster_size] = max_poly_order``. By default,
        for a given cluster orbit, polynomials of order up to the cluster size are
        created. Higher order polynomials are requested either according to cluster
        size using `orbit_branch_max_poly_order` or globally using
        `global_max_poly_order`. The most specific level specified is used.

    Returns
    -------
    clex_basis_specs: casm.bset.cluster_functions.ClexBasisSpecs
        The cluster expansion basis set specifications

    """
    prim = _helpers.as_Prim(prim)

    # cluster specs
    if phenomenal is None:
        generating_group = prim.factor_group
        phenomenal_cluster = None
    elif isinstance(phenomenal, Cluster):
        symgroup_rep = make_integral_site_coordinate_symgroup_rep(
            prim.factor_group.elements, prim.xtal_prim
        )
        generating_group = make_cluster_group(
            cluster=phenomenal,
            group=prim.factor_group,
            lattice=prim.xtal_prim.lattice(),
            integral_site_coordinate_symgroup_rep=symgroup_rep,
        )
        phenomenal_cluster = phenomenal
    elif isinstance(phenomenal, OccEvent):
        symgroup_rep = make_occevent_symgroup_rep(
            prim.factor_group.elements, prim.xtal_prim
        )
        generating_group = make_occevent_group(
            occ_event=phenomenal,
            group=prim.factor_group,
            lattice=prim.xtal_prim.lattice(),
            occevent_symgroup_rep=symgroup_rep,
        )
        phenomenal_cluster = phenomenal.cluster()
    else:
        raise ValueError(
            "Error in build_cluster_functions:"
            "`phenomenal` must be a Cluster, OccEvent, or None"
        )

    cluster_specs = ClusterSpecs(
        xtal_prim=prim.xtal_prim,
        generating_group=generating_group,
        max_length=max_length,
        custom_generators=custom_generators,
        phenomenal=phenomenal_cluster,
        include_phenomenal_sites=False,
        cutoff_radius=cutoff_radius,
    )

    dof_specs = {}
    if occ_site_basis_functions_specs is not None:
        dof_specs["occ"] = {"site_basis_functions": occ_site_basis_functions_specs}

    _orbit_branch_max_poly_order = {}
    if orbit_branch_max_poly_order is not None:
        _orbit_branch_max_poly_order = {
            str(key): value for key, value in orbit_branch_max_poly_order.items()
        }

    return ClexBasisSpecs(
        cluster_specs=cluster_specs,
        basis_function_specs=BasisFunctionSpecs(
            dofs=dofs,
            dof_specs=dof_specs,
            global_max_poly_order=global_max_poly_order,
            orbit_branch_max_poly_order=_orbit_branch_max_poly_order,
        ),
    )


def build_cluster_functions(
    prim: Union[xtal.Prim, casmconfig.Prim, dict, str, pathlib.Path],
    clex_basis_specs: Union[ClexBasisSpecs, dict, str, pathlib.Path],
    prim_neighbor_list: Optional[PrimNeighborList] = None,
    make_equivalents: bool = True,
    make_all_local_basis_sets: bool = True,
    verbose: bool = False,
) -> ClusterFunctionsBuilder:
    """Constructs cluster expansion basis functions

    Parameters
    ----------
    prim: Union[libcasm.xtal.Prim, libcasm.configuration.Prim, dict, str, pathlib.Path]
        The prim, with symmetry information. May be provided as a Prim instance, a Prim
        dict, or the path to a file containing the Prim dict.

    clex_basis_specs: Union[casm.bset.cluster_functions.ClexBasisSpecs, dict, str, \
    pathlib.Path]
        Parameters specifying the cluster orbits and basis function type and order. May
        be provided as a ClexBasisSpecs instance, a ClexBasisSpecs dict, or the path
        to a file containing a ClexBasisSpecs dict.

    prim_neighbor_list: Optional[libcasm.clexulator.PrimNeighborList] = None
        The :class:`PrimNeighborList` is used to uniquely index sites with local
        variables included in the cluster functions, relative to a reference unit
        cell. If not provided, a PrimNeighborList is constructed using default
        parameters that include all sites with degrees of freedom (DoF) and the
        default shape used by CASM projects. In most cases, the default should be
        used.

    make_equivalents: bool = True
        If True, make all equivalent clusters and functions. Otherwise, only
        construct and return the prototype clusters and functions on the prototype
        cluster (i.e. ``i_equiv=0`` only).

    make_all_local_basis_sets: bool = True
        If True, make local clusters and functions for all phenomenal
        clusters in the primitive cell equivalent by prim factor group symmetry.
        Requires that `make_equivalents` is True.

    verbose: bool = False
        Print progress statements

    Returns
    -------
    builder: casm.bset.cluster_functions.ClusterFunctionsBuilder
        The ClusterFunctionsBuilder data structure holds the generated cluster
        functions and associated clusters.

    """
    prim = _helpers.as_Prim(prim)
    prim_neighbor_list = _helpers.as_PrimNeighborList(prim_neighbor_list, prim=prim)

    clex_basis_specs = _helpers.as_ClexBasisSpecs(clex_basis_specs, prim=prim)

    cluster_specs = clex_basis_specs.cluster_specs
    clusters = [orbit[0] for orbit in cluster_specs.make_orbits()]
    phenomenal_cluster = cluster_specs.phenomenal()

    bfunc_specs = clex_basis_specs.basis_function_specs
    orbit_branch_max_poly_order = {
        int(key): value
        for key, value in bfunc_specs.orbit_branch_max_poly_order.items()
    }
    occ_site_basis_functions_specs = None
    if "occ" in bfunc_specs.dof_specs:
        if "site_basis_functions" in bfunc_specs.dof_specs["occ"]:
            occ_site_basis_functions_specs = bfunc_specs.dof_specs["occ"][
                "site_basis_functions"
            ]

    return ClusterFunctionsBuilder(
        prim=prim,
        generating_group=cluster_specs.generating_group(),
        clusters=clusters,
        phenomenal=phenomenal_cluster,
        dofs=bfunc_specs.dofs,
        global_max_poly_order=bfunc_specs.global_max_poly_order,
        orbit_branch_max_poly_order=orbit_branch_max_poly_order,
        occ_site_basis_functions_specs=occ_site_basis_functions_specs,
        prim_neighbor_list=prim_neighbor_list,
        make_equivalents=make_equivalents,
        make_all_local_basis_sets=make_all_local_basis_sets,
        verbose=verbose,
    )


def write_clexulator(
    prim: Union[xtal.Prim, casmconfig.Prim, dict, str, pathlib.Path],
    clex_basis_specs: Union[ClexBasisSpecs, dict, str, pathlib.Path],
    bset_dir: Union[str, pathlib.Path, None] = None,
    prim_neighbor_list: Optional[PrimNeighborList] = None,
    project_name: Optional[str] = None,
    bset_name: str = "default",
    version: str = "v1.basic",
    linear_function_indices: Optional[set[int]] = None,
    cpp_fmt: Optional[CppFormatProperties] = None,
    verbose: bool = True,
    very_verbose: bool = False,
) -> tuple[pathlib.Path, Optional[list[pathlib.Path]], PrimNeighborList]:
    """Write a CASM Clexulator

    Notes
    -----

    The CASM Clexulator is written to the `bset_dir` directory as described in the
    documentation for
    `the CASM Clexulator <https://prisms-center.github.io/CASMcode_pydocs/libcasm/clexulator/2.0/usage/cluster_expansion_details.html#the-casm-clexulator>`_.

    Parameters
    ----------
    prim: Union[libcasm.xtal.Prim, libcasm.configuration.Prim, dict, str, pathlib.Path]
        The prim, with symmetry information. May be provided as a Prim instance, a Prim
        dict, or the path to a file containing the Prim dict.

    clex_basis_specs: Union[casm.bset.cluster_functions.ClexBasisSpecs, dict, str, \
    pathlib.Path]
        Parameters specifying the cluster orbits and basis function type and order. May
        be provided as a ClexBasisSpecs instance, a ClexBasisSpecs dict, or the path
        to a file containing a ClexBasisSpecs dict.

    bset_dir: Union[pathlib.Path, str, None] = None
        The path to the basis set directory where the Clexulator should be written.
        If None, the current working directory is used.

    prim_neighbor_list: Optional[PrimNeighborList] = None
        The :class:`PrimNeighborList` is used to uniquely index sites with local
        variables included in the cluster functions, relative to a reference unit cell.
        If None, a default neighbor list is constructed.

    project_name: Optional[str] = None
        Project name. Used to construct the Clexulator class name. If None, uses
        the prim's title. This must consist of alphanumeric characters and underscores
        only. The first character may not be a number.

    bset_name: str = "default"
        Basis set name. Used to construct the Clexulator class name. This must consist
        of alphanumeric characters and underscores only.

    version: str = "v1.basic"
        The Clexulator version to write. One of:

        - "v1.basic": Standard CASM v1 compatible Clexulator, without automatic
          differentiation
        - "v1.diff": (TODO) CASM v1 compatible Clexulator, with ``fadbad`` automatic
          differentiation enabled

    linear_function_indices: Optional[set[int]] = None
        (Experimental feature) The linear indices of the functions that will be
        included. If None, all functions will be included in the Clexulator. Otherwise,
        only the specified functions will be included in the Clexulator.
        Generally this is not known the first time a Clexulator is generated, but
        after fitting coefficients it may be used to re-generate the Clexulator
        with the subset of the basis functions needed.

    cpp_fmt: Optional[CppFormatProperties] = None
        C++ string formatting properties. If None, default constructor values are used.

    verbose: bool = True
        Print progress statements

    very_verbose: bool = False
        Print detailed progress statements from the cluster functions builder.

    Returns
    -------
    src_path: pathlib.Path
        The path to the Clexulator source file
    local_src_path: Optional[list[pathlib.Path]]
        The paths to the local Clexulator source files
    prim_neighbor_list: libcasm.clexulator.PrimNeighborList
        The PrimNeighborList.
    """
    prim = _helpers.as_Prim(prim)
    clex_basis_specs = _helpers.as_ClexBasisSpecs(clex_basis_specs, prim=prim)
    bset_dir = _helpers.as_bset_dir(bset_dir)
    prim_neighbor_list = _helpers.as_PrimNeighborList(prim_neighbor_list, prim=prim)

    if project_name is None:
        project_name = prim.to_dict().get("title", "")

    writer = ClexulatorWriter(
        bset_dir=bset_dir,
        version=version,
        project_name=project_name,
        bset_name=bset_name,
        linear_function_indices=linear_function_indices,
        cpp_fmt=cpp_fmt,
    )
    writer.write(
        prim=prim,
        clex_basis_specs=clex_basis_specs,
        prim_neighbor_list=prim_neighbor_list,
        verbose=verbose,
        very_verbose=very_verbose,
    )

    return (writer.src_path, writer.local_src_path, prim_neighbor_list)


class _TestSystem:
    """Used by autoconfigure to test writing, compiling, and using a Clexulator"""

    def __init__(
        self,
    ):
        pass

    def __enter__(
        self,
    ):
        import tempfile

        import libcasm.xtal.prims as xtal_prims

        xtal_prim = xtal_prims.FCC(
            r=0.5,
            occ_dof=["A"],
            global_dof=[xtal.DoFSetBasis("Hstrain")],
        )
        self.prim = casmconfig.Prim(xtal_prim)

        clex_basis_specs = make_clex_basis_specs(
            prim=self.prim,
            max_length=[0.0],
            global_max_poly_order=4,
        )

        self.tmp_bset_dir = tempfile.TemporaryDirectory()

        self.src_path, self.local_src_path, self.prim_neighbor_list = write_clexulator(
            prim=self.prim,
            clex_basis_specs=clex_basis_specs,
            bset_dir=self.tmp_bset_dir.name,
            project_name="TestProject",
            bset_name="default",
            version="v1.basic",
        )
        return self

    def try_vars(self, test_vars: dict, verbose: bool = True):
        try:
            from libcasm.casmglobal.__main__ import main as cgmain
        except ImportError:
            raise ImportError("libcasm is not installed")

        import os

        if "CASM_PREFIX" not in test_vars:
            import io
            from contextlib import redirect_stdout

            f = io.StringIO()
            with redirect_stdout(f):
                cgmain(argv=["casmglobal", "--prefix"])
            casm_prefix = f.getvalue().strip()
            os.environ["CASM_PREFIX"] = casm_prefix

        if verbose:
            print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
            print("Trying:")
            if "CASM_PREFIX" not in test_vars:
                print("export CASM_PREFIX=$(python -m libcasm.casmglobal --prefix)")
            for k, v in test_vars.items():
                if v is None:
                    print(f"unset {k}")
                else:
                    print(f'export {k}="{v}"')
            print()

        # set environment variables; keys with value None are removed
        for k, v in test_vars.items():
            if v is None:
                if k in os.environ:
                    del os.environ[k]
            else:
                os.environ[k] = v

        import io
        from contextlib import redirect_stdout

        if verbose:
            print("Make clexulator...")
        self.clexulator = make_clexulator(
            source=str(self.src_path),
            prim_neighbor_list=self.prim_neighbor_list,
        )
        if verbose:
            print("Make clexulator: DONE")

        if verbose:
            print("Test clexulator...")
        if self.clexulator.n_functions() != 22:
            raise RuntimeError("n_functions() != 22")

        self.supercell = casmconfig.Supercell(
            prim=self.prim,
            transformation_matrix_to_super=np.array(
                [
                    [-1, 1, 1],
                    [1, -1, 1],
                    [1, 1, -1],
                ],
                dtype="int",
            ),
        )
        self.supercell_neighbor_list = SuperNeighborList(
            self.supercell.transformation_matrix_to_super,
            self.prim_neighbor_list,
        )

        self.config = casmconfig.Configuration(supercell=self.supercell)
        self.config.set_global_dof_values(
            key="Hstrain", dof_values=np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        )

        self.corr = Correlations(
            supercell_neighbor_list=self.supercell_neighbor_list,
            clexulator=self.clexulator,
            config_dof_values=self.config.dof_values,
        )

        x = self.corr.per_unitcell(self.corr.per_supercell())

        n_func = self.clexulator.n_functions()
        if x.shape != (n_func,):
            raise RuntimeError("correlations shape error")
        if not np.allclose(x, [1.0] + [0.0] * (n_func - 1)):
            raise RuntimeError("correlations value error")
        if verbose:
            print("Test clexulator: DONE")
            print()

    def reset(self):
        import os

        for x in os.listdir(self.tmp_bset_dir.name):
            if x not in ["basis.json", "TestProject_Clexulator_default.cc"]:
                os.remove(os.path.join(self.tmp_bset_dir.name, x))

    def __exit__(self, exc_type, exc_value, traceback):
        self.tmp_bset_dir.cleanup()


def autoconfigure(
    apply_results: bool = True,
    return_results: bool = False,
    user_vars: list[dict] = [],
    verbose: bool = False,
):
    R"""Automatically determine and set environment variables needed for compiling and
    linking Clexulator

    This method attempts to find the environment variables needed to compile and link a
    Clexulator by testing some standard variables sets. The standard variable sets are:

    .. code-block:: Python

        [
            dict(
                CASM_CXXFLAGS=None,
                CASM_SOFLAGS=None,
            ),
            dict(
                CASM_CXXFLAGS="-O3 -Wall -fPIC --std=c++17 -D_GLIBCXX_USE_CXX11_ABI=0 ",
                CASM_SOFLAGS="-shared -Wl,--no-as-needed",
            ),
            dict(
                CASM_CXXFLAGS=None,
                CASM_SOFLAGS="-shared -Wl,--no-as-needed",
            ),
        ]

    The order in which the variable sets are tried may vary depending on the system.
    When "CASM_PREFIX" is not included, then it is assumed to be configured with:

    .. code-block:: bash

        export CASM_PREFIX=$(python -m libcasm.casmglobal --prefix)


    Parameters
    ----------
    apply_results: bool = True
        If True and successful, apply the variables found to the current environment.
        If True and not successful, raise an exception.

    return_results: bool = False
        If True, return configuration results dictionary.

    user_vars: list[dict] = []
        List of dictionaries containing sets of environment variables to test in
        addition to the standard variable sets.

    verbose: bool = True
        If True, print progress statements.

    Returns
    -------
    results: Optional[dict]
        Results of the autoconfiguration tests. Will contains `vars`, a dictionary of
        environment variables that were succesfully used to compile and
        use a Clexulator. If no successful configuration was found, `vars` will be None.
        Will also contain `failed`, a list of dictionaries containing `vars` (failed
        test variables) and `what` (error message).

        Format:

        .. code-block:: Python

            results = {
                "vars": {  # successful variables, or None
                    "CASM_CXXFLAGS": Optional[str],
                    "CASM_SOFLAGS": Optional[str],
                },
                "failed": [  # failed variables
                    {
                        "vars": {  # failed test variables
                            "CASM_CXXFLAGS": Optional[str],
                            "CASM_SOFLAGS": Optional[str],
                        },
                        "what": str,  # description of what failed
                    },
                    ...
                ],
            }

    """

    all_vars = user_vars
    import os
    import subprocess

    if not apply_results:
        orig_environ = dict(os.environ)

    # try to check for clang
    using_clang = False
    cxx = "g++"
    if "CASM_CXX" in os.environ:
        cxx = os.environ["CASM_CXX"]
    elif "CXX" in os.environ:
        cxx = os.environ["CXX"]
    out = subprocess.run(
        [cxx, "--version"], stdout=subprocess.PIPE, encoding="utf-8"
    ).stdout
    if "clang" in out:
        using_clang = True

    # known configuration sets
    set1 = dict(
        CASM_CXXFLAGS=None,
        CASM_SOFLAGS=None,
    )
    set2 = dict(
        CASM_CXXFLAGS="-O3 -Wall -fPIC --std=c++17 -D_GLIBCXX_USE_CXX11_ABI=0 ",
        CASM_SOFLAGS="-shared -Wl,--no-as-needed",
    )
    set3 = dict(
        CASM_CXXFLAGS=None,
        CASM_SOFLAGS="-shared -Wl,--no-as-needed",
    )

    if using_clang:
        all_vars += [set1, set2, set3]
    else:
        all_vars += [set2, set3, set1]

    results = {"vars": None, "failed": []}

    with _TestSystem() as test_system:
        for test_vars in all_vars:
            try:
                test_system.try_vars(test_vars, verbose=verbose)
                results["vars"] = test_vars
                break
            except Exception as e:
                results["failed"].append({"vars": test_vars, "what": str(e)})
                if verbose:
                    print()
            test_system.reset()

    if apply_results is True:
        if results["vars"] is None:
            raise Exception("No successful configuration found")
    else:
        os.environ = orig_environ

    if return_results:
        return results
    else:
        return None



from typing import Optional

from casm.bset.parsing import to_dict


class BasisFunctionSpecs:
    R"""Basis function specifications

    BasisFunctionSpecs is used to specify the type and order of basis functions that
    CASM generates.
    """

    def __init__(
        self,
        dofs: Optional[list[str]] = None,
        dof_specs: dict = {},
        global_max_poly_order: Optional[int] = None,
        orbit_branch_max_poly_order: dict = {},
        param_pack_type: str = "default",
    ):
        """
        .. rubric:: Constructor

        Parameters
        ----------
        dofs: Optional[list[str]] = None
            An array of string of dof type names that should be used to construct basis
            functions. The default value is all DoF types included in the prim.

        dof_specs: dict = {}
            Provides DoF-particular specifications for constructing basis functions,
            as described in the section
            :ref:`DoF Specifications <sec-dof-specifications>`.

        global_max_poly_order: Optional[int] = None
            The maximum polynomial order for cluster functions of continuous DoF.
            If None, the maximum polynomial order is set to the cluster size. May be
            overridden on a per-orbit-branch basis by `orbit_branch_max_poly_order`.

        orbit_branch_max_poly_order: dict = {}
            Specifies the maximum polynomial order on a per-orbit-branch basis, where
            an "orbit branch" is the set of cluster orbits with the same number of
            sites per cluster. Keys indicate the orbit branch, specified using the
            string value of the cluster size as the key, and the value is the maximum
            polynomial order for that orbit branch. For example:

            .. code-block:: Python

                orbit_branch_max_poly_order={
                    "1": 6, # 1-site clusters, max order 6
                    "2": 5, # 2-site clusters, max order 5
                }

        param_pack_type: str = "default"
            Controls the implementation used for evaluating the cluster expansion basis
            functions. Options are “default” or “diff”, which enables fadbad automatic
            differentiating.

        """

        self.dof_specs: dict = dof_specs
        """dict: DoF-particular specifications for constructing basis functions.
        
        See the :class:`casm.bset.BasisFunctionSpecs` constructor for details
        """

        self.dofs: Optional[list[str]] = dofs
        """Optional[list[str]]: An array of string of dof type names that should be
        used to construct basis functions. 
        
        The default value is all DoF types included in the prim.
        """

        self.global_max_poly_order: Optional[int] = global_max_poly_order
        """Optional[int]: Sets default lower bound on maximum polynomial order
        
        For example, if ``global_max_poly_order = 3``, then the maximum polynomial
        order requested for orbit branches not specified in
        :py:data:`~casm.bset.BasisFunctionSpecs.orbit_branch_max_poly_order`
        is ``max(3, cluster_size)``.
        
        See the
        :py:data:`~casm.bset.BasisFunctionSpecs.orbit_branch_max_poly_order`
        documentation for further details.
        """
        self.orbit_branch_max_poly_order: dict = orbit_branch_max_poly_order
        """dict: Sets orbit branch maximum polynomial order
        
        By default, for a given cluster orbit, polynomials of order up to the
        cluster size are created. Higher order polynomials can be requested either
        on a per-orbit-branch or global basis. The most specific level specified
        is used. Orbit branches are specified using the string value of the cluster
        size as a key.
        
        Example:

        .. code-block:: Python
        
            basis_function_specs = BasisFunctionSpecs(
                ...
                # use maximum polynomial order == 7,
                # for orbits of cluster size == 4
                orbit_branch_max_poly_order = {"4": 7},
                
                # use max(3, cluster size), for all other orbits
                global_max_poly_order = 3,
                ...
            )
                
        """

        self.param_pack_type: str = param_pack_type
        """str: Controls the implementation used for evaluating the cluster expansion
        basis functions. 
        
        Options are “default” or “diff”, which enables fadbad automatic differentiating.
        """

    @staticmethod
    def from_dict(data):
        """Construct BasisFunctionSpecs from a Python dictionary

        See the CASM documentation for the `BasisFunctionSpecs format`_.

        .. _`BasisFunctionSpecs format`: https://prisms-center.github.io/CASMcode_docs/formats/casm/basis_set/BasisFunctionSpecs/


        Parameters
        ----------
        data : dict
            A Python dictionary formatted as documented

        Returns
        -------
        basis_function_specs : BasisFunctionSpecs
            The created BasisFunctionSpecs
        """
        return BasisFunctionSpecs(
            dofs=data.get("dofs"),
            dof_specs=data.get("dof_specs", {}),
            global_max_poly_order=data.get("global_max_poly_order"),
            orbit_branch_max_poly_order=data.get("orbit_branch_max_poly_order", {}),
            param_pack_type=data.get("param_pack_type", "default"),
        )

    def to_dict(self):
        """Represent the BasisFunctionSpecs as a Python dictionary

        See the CASM documentation for the `BasisFunctionSpecs format`_.

        .. _`BasisFunctionSpecs format`: https://prisms-center.github.io/CASMcode_docs/formats/casm/basis_set/BasisFunctionSpecs/


        Returns
        -------
        data : dict
            The BasisFunctionSpecs as a Python dictionary

        """
        data = {}
        to_dict(self.dofs, data, "dofs")
        to_dict(self.dof_specs, data, "dof_specs")
        to_dict(self.global_max_poly_order, data, "global_max_poly_order")
        to_dict(self.orbit_branch_max_poly_order, data, "orbit_branch_max_poly_order")

        if self.param_pack_type != "default":
            to_dict(self.param_pack_type, data, "param_pack_type")

        return data



import math

import numpy as np

from casm.bset.cluster_functions import (
    make_orthonormal_discrete_functions,
)


def point_sum(x, phi):
    sum = 0.0
    for i in range(len(x)):
        sum += x[i] * phi[i]
    return sum


def pair_sum(x, phi):
    sum = 0.0
    for i in range(len(x)):
        for j in range(len(x)):
            sum += x[i] * x[j] * phi[i] * phi[j]
    return sum


def triplet_sum(x, phi):
    sum = 0.0
    for i in range(len(x)):
        for j in range(len(x)):
            for k in range(len(x)):
                sum += x[i] * x[j] * x[j] * phi[i] * phi[j] * phi[k]
    return sum


def binary_checks(x, phi_expected):
    abs_tol = 1e-10
    phi = make_orthonormal_discrete_functions(x, abs_tol=abs_tol)
    assert np.allclose(phi, phi_expected, atol=abs_tol)
    assert math.isclose(point_sum(x, phi[1, :]), 0.0, abs_tol=abs_tol)
    assert math.isclose(pair_sum(x, phi[1, :]), 0.0, abs_tol=abs_tol)
    assert math.isclose(triplet_sum(x, phi[1, :]), 0.0, abs_tol=abs_tol)


def test_discrete_functions_1():
    ## binary (checks vs CASM v1)

    x = [1.0, 0.0]
    binary_checks(x, np.array([[1.0, 1.0], [0.0, 1.0]]))

    x = [0.5, 0.5]
    binary_checks(x, np.array([[1.0, 1.0], [-1.0, 1.0]]))

    x = [0.45, 0.55]
    binary_checks(x, np.array([[1.0, 1.0], [1.08386, -0.886796]]))

    x = [0.4, 0.6]
    binary_checks(x, np.array([[1.0, 1.0], [1.13389, -0.755929]]))

    x = [0.3, 0.7]
    binary_checks(x, np.array([[1.0, 1.0], [1.15079, -0.493197]]))

    x = [0.25, 0.75]
    binary_checks(x, np.array([[1.0, 1.0], [1.13389, -0.377964]]))

    x = [0.2, 0.8]
    binary_checks(x, np.array([[1.0, 1.0], [1.1094, -0.27735]]))

    x = [0.1, 0.9]
    binary_checks(x, np.array([[1.0, 1.0], [1.05337, -0.117041]]))

    x = [0.0, 1.0]
    binary_checks(x, np.array([[1.0, 1.0], [1.0, 0.0]]))

    ## ternary

    # composition [1.0, 0.0, 0.0]
    # [0., 1., 0.]
    # [0., 0., 1.]
    phi = make_orthonormal_discrete_functions([1.0, 0.0, 0.0], abs_tol=1e-10)
    assert np.allclose(
        phi, np.array([[1.0, 1.0, 1.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
    )

    # composition [0.0, 1.0, 0.0]
    # [0., 1., 0.]
    # [0., 0., 1.]
    phi = make_orthonormal_discrete_functions([0.0, 1.0, 0.0], abs_tol=1e-10)
    assert np.allclose(
        phi, np.array([[1.0, 1.0, 1.0], [1.0, 0.0, 0.0], [0.0, 0.0, 1.0]])
    )

    # composition [0.0, 0.0, 1.0]
    # [0., 1., 0.]
    # [0., 0., 1.]
    phi = make_orthonormal_discrete_functions([0.0, 0.0, 1.0], abs_tol=1e-10)
    assert np.allclose(
        phi, np.array([[1.0, 1.0, 1.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
    )

    # composition [1/3., 1/3., 1/3.]
    # [-1., 1.]
    phi = make_orthonormal_discrete_functions(
        [1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0], abs_tol=1e-10
    )
    assert np.allclose(
        phi,
        np.array(
            [
                [1.0, 1.0, 1.0],
                [-1.22474487, 0.0, 1.22474487],
                [-0.70710678, 1.41421356, -0.70710678],
            ]
        ),
    )



import argparse
import sys


def _autoconfigure(envfile: str = None, shfile: str = None):
    from casm.bset import autoconfigure

    print("Begin autoconfigure...")
    results = autoconfigure(apply_results=False, return_results=True, verbose=True)

    if results["vars"] is None:
        print("autoconfigure: FAILED")
        for failed in results["failed"]:
            print()
            print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
            print("Tried:")
            print("##")
            if "CASM_PREFIX" not in failed["vars"]:
                print("export CASM_PREFIX=$(python -m libcasm.casmglobal --prefix)")
            for k, v in failed["vars"].items():
                if v is not None:
                    print(f'export {k}="{v}"')
            print("##")
            print()
            print(f'Error: {failed["what"]}')
    else:
        print("autoconfigure: SUCCESS")
        print()
        print("Successful configuration:")
        print()
        print("##")
        if "CASM_PREFIX" not in results["vars"]:
            print("export CASM_PREFIX=$(python -m libcasm.casmglobal --prefix)")
        for k, v in results["vars"].items():
            if v is not None:
                print(f'export {k}="{v}"')
        print("##")

        if envfile:
            with open(envfile, "a") as f:
                if "CASM_PREFIX" not in results["vars"]:
                    import io
                    from contextlib import redirect_stdout

                    from libcasm.casmglobal.__main__ import main as cgmain

                    g = io.StringIO()
                    with redirect_stdout(g):
                        cgmain(argv=["casmglobal", "--prefix"])
                    casm_prefix = g.getvalue().strip()
                    f.write(f"CASM_PREFIX={casm_prefix}\n")
                for k, v in results["vars"].items():
                    if v is not None:
                        f.write(f"{k}={v}\n")

            print()
            print("appended variables to:", envfile)

        if shfile:
            with open(shfile, "w") as f:
                if "CASM_PREFIX" not in results["vars"]:
                    f.write(
                        "export CASM_PREFIX=$(python -m libcasm.casmglobal --prefix)\n"
                    )
                for k, v in results["vars"].items():
                    if v is not None:
                        f.write(f'export {k}="{v}"\n')

            print()
            print(f"wrote: {shfile}")


parser = argparse.ArgumentParser()
parser.add_argument(
    "--autoconfigure",
    action="store_true",
    help="Run autoconfigure to find environment variables for compiling Clexulator.",
)
parser.add_argument(
    "--envfile",
    type=str,
    help="If provided, successful variables are appended to the given file.",
)
parser.add_argument(
    "--shfile",
    type=str,
    help="If provided, a shell script with given name is written that can be used to "
    "set the successful variables.",
)

args = parser.parse_args(args=sys.argv[1:])

if not sys.argv[1:]:
    parser.print_help()
    exit()

if args.autoconfigure:
    _autoconfigure(envfile=args.envfile, shfile=args.shfile)



import os
import pathlib
import shutil
import sys

import numpy as np
import pytest

import libcasm.xtal as xtal

# Pytest will rewrite assertions in test modules, but not elsewhere.
# This tells pytest to also rewrite assertions in utils/helpers.py.
#
pytest.register_assert_rewrite("utils.helpers")


def _win32_longpath(path):
    """
    Helper function to add the long path prefix for Windows, so that shutil.copytree
     won't fail while working with paths with 255+ chars.
    """
    if sys.platform == "win32":
        # The use of os.path.normpath here is necessary since "the "\\?\" prefix
        # to a path string tells the Windows APIs to disable all string parsing
        # and to send the string that follows it straight to the file system".
        # (See https://docs.microsoft.com/pt-br/windows/desktop/FileIO/naming-a-file)
        return "\\\\?\\" + os.path.normpath(path)
    else:
        return path


@pytest.fixture(scope="session")
def session_shared_datadir(tmpdir_factory):
    original_shared_path = pathlib.Path(os.path.realpath(__file__)).parent / "data"
    session_temp_path = tmpdir_factory.mktemp("session_data")
    shutil.copytree(
        _win32_longpath(original_shared_path),
        _win32_longpath(str(session_temp_path)),
        dirs_exist_ok=True,
    )
    return session_temp_path


@pytest.fixture
def lowsym_Hstrain_prim():
    return xtal.Prim(
        lattice=xtal.Lattice(
            np.array(
                [
                    [1.0, 0.3, 0.4],  # a
                    [0.0, 1.2, 0.5],  # b
                    [0.0, 0.0, 1.4],  # c
                ]
            ).transpose()
        ),
        coordinate_frac=np.array(
            [
                [0.0, 0.0, 0.0],
                [0.4, 0.5, 0.6],
                [0.24, 0.25, 0.23],
            ]
        ).transpose(),
        occ_dof=[["A"], ["A"], ["A"]],
        global_dof=[xtal.DoFSetBasis("Hstrain")],
    )


@pytest.fixture
def lowsym_disp_prim():
    return xtal.Prim(
        lattice=xtal.Lattice(
            np.array(
                [
                    [1.0, 0.3, 0.4],  # a
                    [0.0, 1.2, 0.5],  # b
                    [0.0, 0.0, 1.4],  # c
                ]
            ).transpose()
        ),
        coordinate_frac=np.array(
            [
                [0.0, 0.0, 0.0],
                [0.4, 0.5, 0.6],
                [0.24, 0.25, 0.23],
            ]
        ).transpose(),
        occ_dof=[["A"], ["A"], ["A"]],
        local_dof=[
            [xtal.DoFSetBasis("disp")],
            [xtal.DoFSetBasis("disp")],
            [xtal.DoFSetBasis("disp")],
        ],
    )


@pytest.fixture
def lowsym_occ_prim():
    return xtal.Prim(
        lattice=xtal.Lattice(
            np.array(
                [
                    [1.0, 0.3, 0.4],  # a
                    [0.0, 1.2, 0.5],  # b
                    [0.0, 0.0, 1.4],  # c
                ]
            ).transpose()
        ),
        coordinate_frac=np.array(
            [
                [0.0, 0.0, 0.0],
                [0.4, 0.5, 0.6],
                [0.24, 0.25, 0.23],
            ]
        ).transpose(),
        occ_dof=[["A", "B"], ["A", "B"], ["A", "B"]],
    )


@pytest.fixture
def lowsym_Hstrain_disp_prim():
    return xtal.Prim(
        lattice=xtal.Lattice(
            np.array(
                [
                    [1.0, 0.3, 0.4],  # a
                    [0.0, 1.2, 0.5],  # b
                    [0.0, 0.0, 1.4],  # c
                ]
            ).transpose()
        ),
        coordinate_frac=np.array(
            [
                [0.0, 0.0, 0.0],
                [0.4, 0.5, 0.6],
                [0.24, 0.25, 0.23],
            ]
        ).transpose(),
        occ_dof=[["A"], ["A"], ["A"]],
        local_dof=[
            [xtal.DoFSetBasis("disp")],
            [xtal.DoFSetBasis("disp")],
            [xtal.DoFSetBasis("disp")],
        ],
        global_dof=[xtal.DoFSetBasis("Hstrain")],
    )



from typing import Any, Callable, Iterable, Optional

import numpy as np

import libcasm.clusterography as casmclust
import libcasm.configuration as casmconfig
import libcasm.configuration.io as config_io
import libcasm.sym_info as sym_info
import libcasm.xtal as xtal
from casm.bset.cluster_functions._discrete_functions import (
    get_occ_site_functions,
    make_occ_site_functions,
    make_occ_site_functions_info,
)
from casm.bset.cluster_functions._matrix_rep import (
    MakeVariableName,
    OrbitMatrixRepBuilder,
    make_variable_name,
)
from casm.bset.cluster_functions._misc import (
    orbits_to_dict,
)
from casm.bset.polynomial_functions import (
    ExponentSumConstraint,
    FunctionRep,
    PolynomialFunction,
    Variable,
    make_symmetry_adapted_polynomials,
)
from libcasm.clexulator import (
    PrimNeighborList,
    make_default_prim_neighbor_list,
)

from ._clex_basis_specs import (
    ClexBasisSpecs,
)


def _get_dof_types(
    xtal_prim: xtal.Prim,
    dofs: Optional[list[str]] = None,
) -> tuple[list[str], list[str], list[str], dict]:
    """Given a list of dofs, make lists of which are global, local continuous, and local
    discrete.

    Parameters
    ----------
    xtal_prim: xtal.Prim
        The Prim
    dofs: Optional[Iterable[str]] = None
        An iterable of string of dof type names that should be used to construct basis
        functions. The default value is all DoF types included in the prim.

    Returns
    -------
    (global_dof, local_continuous_dof, local_discrete_dof):

        global_dof: list[str]
            The types of global degree of freedom (DoF). All global DoF are treated
            as continuous.

        local_continuous_dof: list[str]
            The types of local discrete degree of freedom (DoF).

        local_discrete_dof: list[str]
            The types of local discrete degree of freedom (DoF).

    """
    all_global_dof = set()
    for _global_dof in xtal_prim.global_dof():
        all_global_dof.add(_global_dof.dofname())

    all_local_continuous_dof = set()
    for _sublattice_dof in xtal_prim.local_dof():
        for _local_dof in _sublattice_dof:
            all_local_continuous_dof.add(_local_dof.dofname())

    # TODO: support other local discrete dof
    all_local_discrete_dof = set()
    for _occ_dof in xtal_prim.occ_dof():
        if len(_occ_dof) > 1:
            all_local_discrete_dof.add("occ")

    if dofs is None:
        global_dof = list(all_global_dof)
        local_continuous_dof = list(all_local_continuous_dof)
        local_discrete_dof = list(all_local_discrete_dof)
    else:
        global_dof = []
        local_continuous_dof = []
        local_discrete_dof = []
        for _dofname in dofs:
            if _dofname in all_global_dof:
                global_dof.append(_dofname)
            elif _dofname in all_local_continuous_dof:
                local_continuous_dof.append(_dofname)
            elif _dofname in all_local_discrete_dof:
                local_discrete_dof.append(_dofname)
            else:
                raise Exception(f"Error in _get_dof_types: Unknown dof '{_dofname}'")
    return (global_dof, local_continuous_dof, local_discrete_dof)


def _default_nlist_sublat_indices(
    xtal_prim: xtal.Prim,
) -> list[int]:
    """Default sublattice indices for PrimNeighborList

    Typically, sublattices are included in the neighbor list if they have allowed DoF,
    either >1 occupant DoF allowed, or 1 continuous DoF.

    Parameters
    ----------
    xtal_prim: xtal.Prim
        The Prim.

    Returns
    -------
    sublat_indices: list[int]
        The indices of the sublattices that should be included in the neighbor list.

    """
    sublat_indices = set()
    occ_dof = xtal_prim.occ_dof()
    for b, sublattice_occ_dof in enumerate(occ_dof):
        if len(sublattice_occ_dof) >= 2:
            sublat_indices.add(b)
    local_dof = xtal_prim.local_dof()
    for b, sublattice_local_dof in enumerate(local_dof):
        if len(sublattice_local_dof) > 0:
            sublat_indices.add(b)
    return list(sublat_indices)


def _make_orbits_data(
    prim: casmconfig.Prim,
    clex_basis_specs: ClexBasisSpecs,
    clusters: list[list[casmclust.Cluster]],
    functions: list[list[list[PolynomialFunction]]],
    coordinate_mode: str,
):
    xtal_prim = prim.xtal_prim
    cluster_specs = clex_basis_specs.cluster_specs
    if xtal_prim != cluster_specs.xtal_prim():
        raise Exception(
            "Error in ClusterFunctionsBuilder.basis_dict: "
            "Prim and cluster specs have different xtal_prim."
        )
    if len(clusters[0][0]) != 0:
        raise Exception(
            "Error in ClusterFunctionsBuilder.basis_dict: "
            "This method requires a null cluster orbit."
        )

    def _brief_desc(op: xtal.SymOp) -> str:
        sym_info = xtal.SymInfo(op, xtal_prim.lattice())
        if coordinate_mode == "cart":
            return sym_info.brief_cart()
        elif coordinate_mode == "frac":
            return sym_info.brief_frac()
        else:
            raise Exception(
                "Error in ClusterFunctionsBuilder.basis_dict: "
                f"Unknown coordinate_mode: {coordinate_mode}, "
                "must be 'cart' or 'frac'."
            )

    is_periodic = cluster_specs.phenomenal() is None
    generating_group_site_rep = casmclust.make_integral_site_coordinate_symgroup_rep(
        group_elements=cluster_specs.generating_group().elements,
        xtal_prim=prim.xtal_prim,
    )

    orbits_data = []
    linear_function_index = 0
    for i_orbit, orbit_functions in enumerate(functions):
        orbit = clusters[i_orbit]
        prototype = orbit[0]

        if is_periodic:
            cluster_group = casmclust.make_cluster_group(
                cluster=prototype,
                group=cluster_specs.generating_group(),
                lattice=xtal_prim.lattice(),
                integral_site_coordinate_symgroup_rep=generating_group_site_rep,
            )
        else:
            cluster_group = casmclust.make_local_cluster_group(
                cluster=prototype,
                phenomenal_group=cluster_specs.generating_group(),
                integral_site_coordinate_symgroup_rep=generating_group_site_rep,
            )

        orbit_data = {
            "linear_orbit_index": i_orbit,
            "mult": len(orbit),
            "prototype": orbit[0].to_dict(
                xtal_prim=xtal_prim, phenomenal=cluster_specs.phenomenal()
            ),
        }
        orbit_data["prototype"]["invariant_group"] = cluster_group.head_group_index
        orbit_data["prototype"]["invariant_group_descriptions"] = [
            _brief_desc(op) for op in cluster_group.elements
        ]

        cluster_functions = []
        if len(prototype) == 0:
            # add constant function
            cluster_functions.append(
                {
                    "\\Phi_{0}": "1",
                    "linear_function_index": linear_function_index,
                }
            )
            linear_function_index += 1
        if len(orbit_functions) != 0:
            for i_func, func in enumerate(orbit_functions[0]):
                cluster_functions.append(
                    {
                        f"\\Phi_{{{linear_function_index}}}": func.latex_formula(),
                        "linear_function_index": linear_function_index,
                    }
                )
                linear_function_index += 1
        orbit_data["cluster_functions"] = cluster_functions
        orbits_data.append(orbit_data)
    return orbits_data


def _make_occ_site_function_basis_data(
    prim: casmconfig.Prim,
    occ_site_function: dict,
):
    """Generate basis dict for an occupation site basis function

    Parameters
    ----------
    prim: casmconfig.Prim
        The Prim
    occ_site_function: dict
        Occupation site basis function values, must include:

        - `"sublattice_index"`: int, index of the sublattice
        - `"value"`: list[list[float]], list of the site basis function values,
          as ``value[function_index][occupant_index]``.


    Returns
    -------
    data: dict
        A dict with the format:

        .. code-block:: Python

            # b: sublattice index,
            # r: site basis function index
            # occ1, occ2, etc.: Occupant DoF names, from prim.xtal_prim.occ_dof()[b]
            {
                "basis": {
                    "\\phi_{b,f}": {
                        "occ1": value1,
                        "occ2": value2,
                        ...
                    },
                    ...
                },
                "value": [
                    [value1, value2, ...],
                    ...
                ]
            }
    """
    b = occ_site_function["sublattice_index"]
    phi = np.array(occ_site_function["value"])
    occ_dof = prim.xtal_prim.occ_dof()[b]
    if len(occ_dof) != phi.shape[0] or len(occ_dof) != phi.shape[1]:
        raise Exception(
            "Error in _make_occ_site_function_basis_data: "
            "Inconsistent number of site basis functions or occupants. "
            "Sublattice index: {b}; Phi shape = {phi.shape}; "
        )
    data = {
        "basis": {},
        "value": phi.tolist(),
    }
    for r in range(phi.shape[0]):
        key = f"\\phi_{{{b},{r}}}"
        d = {}
        for c in range(phi.shape[1]):
            d[occ_dof[c]] = phi[r][c]
        data["basis"][key] = d
    return data


def _make_site_functions_data(
    prim: casmconfig.Prim,
    occ_site_functions: Optional[list[dict]],
):
    """Generate basis dict for site functions

    Parameters
    ----------
    prim: casmconfig.Prim
        The Prim
    occ_site_functions: Optional[list[dict]]
        List of occupation site basis functions. For each sublattice with discrete
        site basis functions, must include:

        - `"sublattice_index"`: int, index of the sublattice
        - `"value"`: list[list[float]], list of the site basis function values,
          as ``value[function_index][occupant_index]``.

    Returns
    -------
    data: list[dict]
        List of site functions data, with the format:

        .. code-block:: Python

            [
                {
                    "sublat": b,  # sublattice index
                    "asym_unit": a,  # asymmetric unit index
                    "occ": { # if discrete site basis functions are present
                        "basis": {
                            "\\phi_{b,f}": {
                                "occ1": value1,
                                "occ2": value2,
                                ...
                            },
                            ...
                        },
                        "value": [
                            [value1, value2, ...],
                            ...
                        ]
                    }
                },
                ...
            ]

    """

    sublat_to_asym_unit = {}
    for a, sublat_indices in enumerate(xtal.asymmetric_unit_indices(prim.xtal_prim)):
        for b in sublat_indices:
            sublat_to_asym_unit[b] = a

    site_functions_data = [
        {"sublat": b, "asym_unit": a} for b, a in sublat_to_asym_unit.items()
    ]

    if occ_site_functions is not None:
        for occ_site_function in occ_site_functions:
            b = occ_site_function["sublattice_index"]
            site_functions_data[b]["occ"] = _make_occ_site_function_basis_data(
                prim=prim, occ_site_function=occ_site_function
            )
    return site_functions_data


def update_variables(
    prim: casmconfig.Prim,
    cluster: casmclust.Cluster,
    function: PolynomialFunction,
    prim_neighbor_list: PrimNeighborList,
    translation: Optional[np.ndarray] = None,
    make_variable_name_f: Optional[Callable] = None,
    local_discrete_dof: Optional[list[str]] = None,
):
    """Update variables after transforming cluster function to assign
    neighborhood_site_index and name for function variables

    Parameters
    ----------
    prim: libcasm.configuration.Prim
        The prim
    cluster: libcasm.clusterography.Cluster
        The cluster associated with the function.
    function: casm.bset.polynomial_functions.PolynomialFunction
        A PolynomialFunction with variables that have cluster_site_index set
        referring to which site in `cluster` they are associated with.
    prim_neighbor_list: libcasm.clexulator.PrimNeighborList
        The neighbor list
    translation: Optional[np.ndarray],
        Optional translation to apply to cluster.
    make_variable_name_f: Optional[Callable] = None
        Allows specifying a custom class to construct variable names. The default
        class used is :class:`~casm.bset.cluster_functions.MakeVariableName`.
        Custom classes should have the same `__call__` signature as
        :class:`~casm.bset.cluster_functions.MakeVariableName`, and have
        `occ_var_name` and `occ_var_desc` attributes.
    local_discrete_dof: Optional[list[str]] = None
        The types of local discrete degree of freedom (DoF).
    """
    if make_variable_name_f is None:
        make_variable_name_f = make_variable_name
    if local_discrete_dof is None:
        local_discrete_dof = list()

    for var in function.variables:
        if var.cluster_site_index is not None:
            integral_site_coordinate = cluster[var.cluster_site_index]
            if translation is not None:
                integral_site_coordinate = integral_site_coordinate + translation

            # Set neighborhood site index
            var.neighborhood_site_index = prim_neighbor_list.neighbor_index(
                integral_site_coordinate
            )

            # Set variable name:
            var.name = make_variable_name_f(
                xtal_prim=prim.xtal_prim,
                key=var.key,
                site_basis_function_index=var.site_basis_function_index,
                component_index=var.component_index,
                cluster_site_index=var.cluster_site_index,
                sublattice_index=integral_site_coordinate.sublattice(),
                local_discrete_dof=local_discrete_dof,
            )


def make_equivalent_cluster_basis_sets(
    prototype_cluster_basis_set: list[list[PolynomialFunction]],
    equivalence_map_clusters: list[list[casmclust.Cluster]],
    equivalence_map_inv_matrix_rep: list[list[np.ndarray]],
    prim_neighbor_list: PrimNeighborList,
    verbose: bool = False,
    prim: Optional[casmconfig.Prim] = None,
    i_orbit: Optional[int] = None,
    make_variable_name_f: Optional[Callable] = None,
    local_discrete_dof: Optional[list[str]] = None,
) -> tuple[list[casmclust.Cluster], list[list[PolynomialFunction]]]:
    orbit = []
    orbit_basis_sets = []

    # generate basis sets on all equivalent clusters
    if verbose:
        print("Generating equivalents...")
        print()
    # i_equiv: index for equivalent clusters
    # M_list: sym rep matrices (1 for each sym op that prototype cluster to
    #     the equivalent cluster, only the first is needed here)
    for i_equiv, M_list in enumerate(equivalence_map_inv_matrix_rep):
        # add equivalent cluster to orbit of clusters
        equiv_cluster = equivalence_map_clusters[i_equiv][0]
        orbit.append(equiv_cluster)
        if verbose:
            print(f"~~~ i_orbit: {i_orbit}, i_equiv: {i_equiv} ~~~")
            print("Equivalent cluster:")
            print(xtal.pretty_json(equiv_cluster.to_dict(prim.xtal_prim)))
            print()

        # add basis set for equivalent cluster to orbit of basis sets
        equiv_basis_set = []
        for f_prototype in prototype_cluster_basis_set:
            M = M_list[0]
            S = FunctionRep(matrix_rep=M)
            f_equiv = S * f_prototype
            update_variables(
                prim=prim,
                cluster=equiv_cluster,
                function=f_equiv,
                prim_neighbor_list=prim_neighbor_list,
                make_variable_name_f=make_variable_name_f,
                local_discrete_dof=local_discrete_dof,
            )
            assert f_equiv.variables is not f_prototype.variables
            for i in range(len(f_prototype.variables)):
                assert f_equiv.variables[i] is not f_prototype.variables[i]
            equiv_basis_set.append(f_equiv)
        orbit_basis_sets.append(equiv_basis_set)
        if verbose:
            print("Equivalent cluster basis set:")
            if len(equiv_basis_set) == 0:
                print("Empty")
            for i_func, func in enumerate(equiv_basis_set):
                print(f"~~~ order: {func.order()}, function_index: {i_func} ~~~")
                func._basic_print()
                print()
            print()

    return (orbit, orbit_basis_sets)


def make_constraints(
    prototype_cluster: casmclust.Cluster,
    prototype_variables: list[Variable],
    prototype_variable_subsets: list[set[int]],
    occ_site_functions: list[dict],
    local_discrete_dof: list[str],
):
    """Make constraints for constructing symmetry adapted polynomials

    Parameters
    ----------
    prototype_cluster: libcasm.clusterography.Cluster
        The prototype cluster.
    prototype_variables: list[casm.bset.polynomial_functions.Variable]
        The prototype cluster function variables, including global and local variables.
    prototype_variable_subsets: list[set[int]]
        The indices of Variable in `prototype_variables` which mix under application of
        symmetry, or permute as a group. A subset could be strain variables,
        displacement variables on a site, or occupant site basis functions on a site.
    occ_site_functions: list[dict]
        List of occupation site basis functions. For each sublattice with discrete
        site basis functions, must include:

        - `"sublattice_index"`: int, index of the sublattice
        - `"value"`: list[list[float]], list of the site basis function values,
          as ``value[function_index][occupant_index]``.

    local_discrete_dof: list[str]
            The types of local discrete degree of freedom (DoF).

    Returns
    -------
    constraints: list[casm.bset.polynomial_functions.ExponentSumConstraint]
        Constraints that set the exponent for constant discrete site functions to 0,
        and ensure one and only one of mutually exclusive discrete variables
        is included in any monomial.
    """
    constraints = []
    for subset in prototype_variable_subsets:
        var = prototype_variables[subset[0]]
        if var.cluster_site_index is None or var.key not in local_discrete_dof:
            continue

        # Find the constant site function and set constraint to exclude
        found_constant_function = False
        b = None
        for i_var in subset:
            var = prototype_variables[i_var]
            b = prototype_cluster[var.cluster_site_index].sublattice()
            site_function_index = var.site_basis_function_index
            phi = get_occ_site_functions(
                occ_site_functions=occ_site_functions,
                sublattice_index=b,
                site_function_index=site_function_index,
            )

            if (phi == 1.0).all():
                constraints.append(ExponentSumConstraint(variables=[i_var], sum=[0]))
                found_constant_function = True
        if not found_constant_function:
            raise Exception(
                "Error in make_constraints: "
                f"discrete DoF (sublattice={b}) has no constant site basis function"
            )

        # Set mutually exclusive discrete variable subset constraint
        constraints.append(ExponentSumConstraint(variables=subset, sum=[0, 1]))

    return constraints


class ClusterFunctionsBuilder:
    """Constructs cluster functions

    Notes
    -----

    Cluster functions are generated using the following steps:

    1. A cluster from each orbit is input, along with the phenomenal cluster if local
       cluster functions are to be generated.

    2. For each orbit,
       :func:`~casm.bset.polynomial_functions.make_symmetry_adapted_polynomials`
       is used to construct symmetry-adapted polynomial functions
       (:class:`~casm.bset.polynomial_functions.PolynomialFunction`) using matrix
       representations constructed by
       :class:`~casm.bset.cluster_functions.OrbitMatrixRepBuilder`.

       - For functions of discrete degrees of freedom (DoF), the polynomial functions
         include a single occupation site basis function on each site in a cluster.
       - For functions that include continuous DoF, by default, polynomials of order up
         to the cluster size are created. Higher order polynomials are requested either
         according to cluster size using the `orbit_branch_max_poly_order` parameter or
         globally using `global_max_poly_order`. For each orbit, the most specific
         level specified is used.
       - Functions are generated for all clusters in the orbit by default. If only
         functions on the prototype cluster are needed, then the `make_equivalents`
         parameter can be set to `False`.
       - The `prim_neighbor_list` is used to set the `neighborhood_site_index` for the
         :class:`~casm.bset.polynomial_functions.Variable` corresponding to local DoF
         used by each :class:`~casm.bset.polynomial_functions.PolynomialFunction`.

    3. If this is a local cluster expansion (if `phenomenal` is not None), then
       local cluster functions are generated for the orbits associated with each of the
       symmetrically equivalent phenomenal clusters.

       - The symmetrically equivalent, but distinct, local cluster expansions are
         found by using the prim factor group to generate equivalent phenomenal
         clusters and excluding those that are equivalent by `generating_group`
         symmetry.
       - For local cluster expansions, all local cluster expansions in the orbit are
         constructed by default. If only functions in the original orbits are needed,
         then the `make_all_local_basis_sets` parameter can be set to `False`.

    """

    def __init__(
        self,
        prim: casmconfig.Prim,
        generating_group: sym_info.SymGroup,
        clusters: Optional[Iterable[casmclust.Cluster]] = None,
        phenomenal: Optional[casmclust.Cluster] = None,
        dofs: Optional[Iterable[str]] = None,
        global_max_poly_order: Optional[int] = None,
        orbit_branch_max_poly_order: Optional[dict[int, int]] = None,
        occ_site_basis_functions_specs: Any = None,
        prim_neighbor_list: Optional[PrimNeighborList] = None,
        make_equivalents: bool = True,
        make_all_local_basis_sets: bool = True,
        make_variable_name_f: Optional[Callable] = None,
        verbose: bool = False,
    ):
        """

        .. rubric:: Constructor

        Parameters
        ----------
        prim: libcasm.configuration.Prim
            The Prim, with symmetry information
        generating_group: libcasm.sym_info.SymGroup
            The symmetry group for generating cluster functions. For periodic cluster
            functions, this is the prim factor group (usually) or a subgroup. For
            local cluster functions, this is the cluster group of the phenomenal
            cluster or a subgroup (often the subgroup which leaves an
            :class:`~libcasm.occ_events.OccEvent` invariant).
        clusters: Iterable[libcasm.clusterography.Cluster]
            An iterable of :class:`~libcasm.clusterography.Cluster` containing the
            a single cluster from each orbit to generate functions for. If not provided,
            clusters are generated using the `max_length`, phenomenal`, and
            `cutoff_radius` parameters.
        phenomenal: Optional[libcasm.clusterography.Cluster] = None
            For local cluster functions, specifies the sites about which
            local-clusters orbits are generated. The phenomenal cluster must be chosen
            from one of the equivalents that is generated by
            :func:`~libcasm.clusterography.make_periodic_orbit` using the prim factor
            group.
        dofs: Optional[Iterable[str]] = None
            An iterable of string of dof type names that should be used to construct
            basis functions. The default value is all DoF types included in the prim.
        global_max_poly_order: Optional[int] = None
            The maximum order of polynomials of continuous DoF to generate, for any
            orbit not specified more specifically by `orbit_branch_max_poly_order`.
        orbit_branch_max_poly_order: Optional[dict[int, int]] = None
            Specifies for continuous DoF the maximum polynomial order to generate by
            cluster size, according to
            ``orbit_branch_max_poly_order[cluster_size] = max_poly_order``. By default,
            for a given cluster orbit, polynomials of order up to the cluster size are
            created. Higher order polynomials are requested either according to cluster
            size using `orbit_branch_max_poly_order` or globally using
            `global_max_poly_order`. The most specific level specified is used.
        occ_site_basis_functions_specs: Any = None
            Provides instructions for constructing occupation site basis functions.
            The most common options are "chebychev" or "occupation". This
            parameter corresponds to the value of

            .. code-block:: Python

                "dof_specs": {
                    "occ": {
                        "site_basis_functions": ...
                    }
                }

            as described in detail in the section
            :ref:`DoF Specifications <sec-dof-specifications>` and is required for
            functions of occupation DoF.
        prim_neighbor_list: Optional[libcasm.clexulator.PrimNeighborList] = None
            The :class:`PrimNeighborList` is used to uniquely index sites with local
            variables included in the cluster functions, relative to a reference unit
            cell. If not provided, a PrimNeighborList is constructed using default
            parameters that include all sites with degrees of freedom (DoF) and the
            default shape used by CASM projects. In most cases, the default should be
            used.
        make_equivalents: bool = True
            If True, make all equivalent clusters and functions. Otherwise, only
            construct functions on the prototype cluster (i.e. ``i_equiv=0`` only).
        make_all_local_basis_sets: bool = True
            If True, make local clusters and functions for all phenomenal
            clusters in the primitive cell equivalent by prim factor group symmetry.
            Requires that `make_equivalents` is True.
        make_variable_name_f: Optional[Callable] = None
            Allows specifying a custom class to construct variable names. The default
            class used is :class:`~casm.bset.cluster_functions.MakeVariableName`.
            Custom classes should have the same `__call__` signature as
            :class:`~casm.bset.cluster_functions.MakeVariableName`, and have
            `occ_var_name` and `occ_var_desc` attributes.
        verbose: bool = False
            Print progress statements

        """

        if orbit_branch_max_poly_order is None:
            orbit_branch_max_poly_order = {}

        if prim_neighbor_list is None:
            prim_neighbor_list = make_default_prim_neighbor_list(
                xtal_prim=prim.xtal_prim
            )

        global_dof, local_continuous_dof, local_discrete_dof = _get_dof_types(
            prim.xtal_prim, dofs
        )

        only_discrete = len(global_dof) + len(local_continuous_dof) == 0

        # Validation
        if len(local_discrete_dof) == 0:
            if occ_site_basis_functions_specs is not None:
                print(
                    "\n"
                    "****************************************************************\n"
                    "** Warning in ClusterFunctionsBuilder:                        **\n"
                    "** No local discrete DoF, but:                                **\n"
                    "** - occ_site_basis_functions_specs is not None               **\n"
                    "** Will use occ_site_basis_functions_specs=None               **\n"
                    "****************************************************************\n"
                )
                occ_site_basis_functions_specs = None
        if len(global_dof + local_continuous_dof) > 0:
            if global_max_poly_order is None and len(orbit_branch_max_poly_order) == 0:
                print(
                    "\n"
                    "****************************************************************\n"
                    "** Warning in ClusterFunctionsBuilder:                        **\n"
                    "** Continuous DoF included, but:                              **\n"
                    "** - global_max_poly_order is None, and                       **\n"
                    "** - len(orbit_branch_max_poly_order) == 0                    **\n"
                    "** Did you forget to set global_max_poly_order??              **\n"
                    "****************************************************************\n"
                )

        # Data
        self._prim = prim
        """libcasm.configuration.Prim: The Prim, with symmetry information"""

        self._generating_group = generating_group
        """libcasm.sym_info.SymGroup: The symmetry group used to generate clusters and \
        functions."""

        self.global_dof = global_dof
        """list[str]: List of global continuous DoF included in functions. """

        self.local_continuous_dof = local_continuous_dof
        """list[str]: List of local continuous DoF included in functions."""

        self.local_discrete_dof = local_discrete_dof
        """list[str]: List of local discrete DoF included in functions."""

        local_dof = local_continuous_dof + local_discrete_dof
        self.local_dof = local_dof
        """list[str]: The types of local continuous and discrete degrees of freedom \
        (DoF) included in the matrix representation.
        """

        self.only_discrete = only_discrete
        """bool: True if only discrete DoF are included in functions."""

        self._clusters = clusters
        """list[libcasm.clusterography.Cluster]: A list of clusters with a single \
        cluster from each orbit to generate functions for.
        """

        self._phenomenal = phenomenal
        """Optional[Cluster]: If provided, specifies the sites about which \
        local-clusters orbits are generated. 
        
        The phenomenal cluster must be chosen from one of the equivalents that is 
        generated by :func:`~libcasm.clusterography.make_periodic_orbit` using the prim
        factor group.
        """

        self._global_max_poly_order = global_max_poly_order
        """int: The default maximum polynomial order.
        
        The maximum order of polynomials of continuous DoF to generate, for any orbit
        not specified more specifically by `orbit_branch_max_poly_order`.
        """

        self._orbit_branch_max_poly_order = orbit_branch_max_poly_order
        """dict[int, int]: The per-orbit-branch maximum polynomial order.
        
        Specifies for continuous DoF the maximum polynomial order to generate by
        cluster size, according to
        ``orbit_branch_max_poly_order[cluster_size] = max_poly_order``. By default, for
        a given cluster orbit, polynomials of order up to the cluster size are created.
        Higher order polynomials are requested either according to cluster size
        using `orbit_branch_max_poly_order` or globally using `global_max_poly_order`.
        The most specific level specified is used.
        """

        self._occ_site_basis_functions_specs = occ_site_basis_functions_specs
        """Union[str, list[dict], None]: Instructions for constructing occupation \
        site basis functions.
            
        The accepted options are "chebychev", "occupation", or a `list[dict]`
        a specifying sublattice-specific choice of site basis functions. 
        
        This parameter corresponds to the value of

        .. code-block:: Python

            "dof_specs": {
                "occ": {
                    "site_basis_functions": ...
                }
            }

        as described in detail in the section
        :ref:`DoF Specifications <sec-dof-specifications>` and is required for
        functions of occupation DoF."""

        occ_site_functions = []
        if occ_site_basis_functions_specs is not None:
            occ_site_functions = make_occ_site_functions(
                prim=prim,
                occ_site_basis_functions_specs=occ_site_basis_functions_specs,
            )

        self.occ_site_functions = occ_site_functions
        """list[dict]: List of occupation site basis functions.
        
        For each sublattice with discrete site basis functions, must include:

        - `"sublattice_index"`: int, index of the sublattice
        - `"value"`: list[list[float]], list of the site basis function values,
          as ``value[function_index][occupant_index]``.
       
        """

        occ_site_functions_info = make_occ_site_functions_info(
            prim=prim,
            occ_site_functions=occ_site_functions,
        )
        # note: this generates default occ_var_name and occ_var_desc, which may be
        # overridden by a custom make_variable_name_f
        self.occ_site_functions_info = occ_site_functions_info
        """dict: Information about occupation site basis functions.
        
        Occupation site basis functions info, with format:

        - `"max_function_index"`: int, The maximum site function index, across all
          sublattices.
        - `"all_sublattices_have_same_site_functions"`: bool, True if all _sublattices
          have same site functions; False otherwise.
        - `"occ_var_name"`: str, A variable name template for the site functions,
          which may be formated using `b` for sublattice index and `m` for site function
          index (i..e ``occ_var_name.format(b=0, m=1)``).
        - `"occ_var_desc"`: str, A description of the occupation
          variable, including a description of the subscript indices.
        """

        self.prim_neighbor_list = prim_neighbor_list
        """libcasm.clexulator.PrimNeighborList: The PrimNeighborList, expanded as \
        necessary
        
        The :class:`~libcasm.clexulator.PrimNeighborList` used to uniquely index sites 
        with local variables included in the cluster functions, relative to a reference 
        unit cell.
        """

        self._make_equivalents = make_equivalents
        """bool: If True, make all equivalent clusters and functions. 
        
        Otherwise, only construct functions on the prototype cluster.
        """

        self._make_all_local_basis_sets = make_all_local_basis_sets
        """bool: If True, make all equivalent orbits of local clusters and local \
        functions. 

        If True, make local clusters and functions for all phenomenal clusters in the 
        primitive cell equivalent by prim factor group symmetry. Requires that 
        `make_equivalents` is True.
        """

        if make_variable_name_f is None:
            make_variable_name_f = MakeVariableName(
                occ_var_name=self.occ_site_functions_info.get("occ_var_name"),
                occ_var_desc=self.occ_site_functions_info.get("occ_var_desc"),
            )
        else:
            self.occ_site_functions_info[
                "occ_var_name"
            ] = make_variable_name_f.occ_var_name
            self.occ_site_functions_info[
                "occ_var_desc"
            ] = make_variable_name_f.occ_var_desc
        self._make_variable_name_f = make_variable_name_f
        """Callable: Function used to construct variable names.
        
        Allows specifying a custom class to construct variable names. The default
        class used is :class:`MakeVariableName`. Custom classes should have
        the same `__call__` signature as :class:`MakeVariableName`, have
        `occ_var_name`, `occ_var_desc`, `occ_var_indices` attributes, and a `to_dict`
        method.
        """

        self._verbose = verbose
        """bool: If True, print progress statements"""

        ## Construct cluster functions on a prototype cluster for each orbit
        prototype_basis_sets = []
        orbit_matrix_rep_builders = []
        constraints = []
        for i_orbit, cluster in enumerate(self._clusters):
            _1, _2, _3 = self._build_prototype_basis_set(i_orbit, cluster)
            prototype_basis_sets.append(_1)
            orbit_matrix_rep_builders.append(_2)
            constraints.append(_3)
        self.prototype_basis_sets = prototype_basis_sets
        """list[list[PolynomialFunction]]: For each orbit, the symmetry adapted \
        polynomial cluster functions for the prototype cluster.
        
        The function ``prototype_basis_sets[i_orbit][i_func]`` is the `i_func`-th 
        function on the prototype cluster in the `i_orbit`-th orbit.
        
        The orbits are generated from the input clusters (stored in 
        :py:data:`~casm.bset.cluster_functions.ClusterFunctionBuilder._clusters`), and 
        stored in the same order. The prototype cluster is determined from sorting the 
        equivalent clusters in the orbit and may not be the input cluster.
        """

        self.orbit_matrix_rep_builders = orbit_matrix_rep_builders
        """list[casm.bset.cluster_functions.OrbitMatrixRepBuilder]: For each orbit,
        the OrbitMatrixRepBuilder contains matrix representations of symmetry
        operations for generating symmetry-adapted polynomial cluster functions on the
        orbit prototype, coupling all local and global DoFs, and matrix representations
        for constructing cluster functions on symmetrically equivalent clusters.
        """

        self.constraints = constraints
        """list[list[ExponentSumConstraint]]: For each orbit, the constraints used \
        when constructing `prototype_basis_set` to ensure one and only one of \
        mutually exclusive discrete variables is included in any monomial.
         
        If a local cluster expansion, ``constraints[i_orbit]`` are also the correct
        constraints for the `i_orbit`-th orbit about each equivalent phenomenal cluster.
        """

        ## Construct equivalent cluster functions
        orbit_basis_sets = None
        orbit_clusters = None
        equivalent_orbit_basis_sets = None
        equivalent_orbit_clusters = None
        if make_equivalents:
            orbit_basis_sets = []
            orbit_clusters = []

            ## Transform cluster functions
            # from prototype clusters to equivalent clusters
            for i_orbit, orbit_matrix_rep_builder in enumerate(
                self.orbit_matrix_rep_builders
            ):
                _1, _2 = self._build_orbit_basis_sets(
                    i_orbit=i_orbit,
                    orbit_matrix_rep_builder=orbit_matrix_rep_builder,
                    prototype_basis_set=self.prototype_basis_sets[i_orbit],
                )
                orbit_basis_sets.append(_1)
                orbit_clusters.append(_2)

                ## For local cluster expansions, transform cluster functions
                # from the local basis set around the original phenomenal cluster
                # to the local basis sets around equivalent phenomenal clusters
                if self._phenomenal is not None and make_all_local_basis_sets:
                    _3, _4 = self._build_equivalent_orbit_basis_sets(
                        i_orbit=i_orbit,
                        orbit_matrix_rep_builder=orbit_matrix_rep_builder,
                        orbit_basis_sets=_1,
                        orbit_clusters=_2,
                    )
                    n_clex = len(_3)
                    if equivalent_orbit_basis_sets is None:
                        equivalent_orbit_basis_sets = [list() for _ in range(n_clex)]
                        equivalent_orbit_clusters = [list() for _ in range(n_clex)]
                    for i_clex in range(n_clex):
                        equivalent_orbit_basis_sets[i_clex].append(_3[i_clex])
                        equivalent_orbit_clusters[i_clex].append(_4[i_clex])

        self.functions = orbit_basis_sets
        """list[list[list[PolynomialFunction]]]: The generated cluster functions

        Polynomial functions, where ``functions[i_orbit][i_equiv][i_func]``, is the 
        `i_func`-th function on the cluster given by `clusters[i_orbit][i_equiv]`.
        """

        n_functions = 1
        for orbit_basis_set in orbit_basis_sets:
            n_functions += len(orbit_basis_set[0])
        self.n_functions = n_functions
        """int: The total number of symmetrically distinct cluster functions generated
        for all clusters in all orbits plus one (for the constant term)."""

        self.clusters = orbit_clusters
        """list[list[libcasm.clusterography.Cluster]]: The clusters for which 
        cluster functions have been constructed
        
        The cluster ``clusters[i_orbit][i_equiv]`` is the `i_equiv`-th symmetrically 
        equivalent cluster in the `i_orbit`-th orbit.

        The order of sites in the clusters is not arbitrary, it is consistent with the 
        `cluster_site_index` of the :class:`~casm.bset.polynomial_functions.Variable`
        used in the :class:`~casm.bset.polynomial_functions.PolynomialFunction` stored 
        in :py:data:`~ClusterFunctionsBuilder.functions`.
        """

        self.equivalent_functions = equivalent_orbit_basis_sets
        """Optional[list[list[list[list[\
        casm.bset.polynomial_functions.PolynomialFunction]]]]]: \
        The generated cluster functions about all equivalent phenomenal clusters (if \
        a local cluster expansion).
        
        Symmetry adapted polynomial cluster functions for each cluster in each
        orbit about a symmetrically equivalent phenomenal cluster,
        where ``equivalent_functions[i_clex][i_orbit][i_equiv][i_func]``, is the
        `i_func`-th function on the `i_equiv`-th cluster in the `i_orbit`-th orbit about
        the `i_clex`-th equivalent phenomenal cluster.
        """

        self.equivalent_clusters = equivalent_orbit_clusters
        """Optional[list[list[list[libcasm.clusterography.Cluster]]]]: Orbit of \
        clusters for which cluster functions have been constructed about all \
        equivalent phenomenal clusters (if a local cluster expansion).

        The cluster ``equivalent_clusters[i_clex][i_orbit][i_equiv]`` is the
        `i_equiv`-th symmetrically equivalent cluster in the `i_orbit`-th orbit about 
        the `i_clex`-th equivalent phenomenal cluster.

        The order of sites in the clusters is not arbitrary, it is consistent with the 
        `cluster_site_index` of the :class:`~casm.bset.polynomial_functions.Variable`
        used in the :class:`~casm.bset.polynomial_functions.PolynomialFunction` stored 
        in :py:data:`~ClusterFunctionsBuilder.equivalent_functions`.
        """

    def to_dict(self):
        data = {}

        def _clusters_data(clusters):
            clusters_data = []
            for orbit_clusters in clusters:
                orbit_clusters_data = []
                for cluster in orbit_clusters:
                    orbit_clusters_data.append({"sites": cluster.to_list()})
                clusters_data.append(orbit_clusters_data)
            return clusters_data

        def _functions_data(functions):
            functions_data = []
            for orbit_functions in functions:
                orbit_functions_data = []
                for equiv_functions in orbit_functions:
                    equiv_functions_data = []
                    for function in equiv_functions:
                        equiv_functions_data.append(function.to_dict())
                    orbit_functions_data.append(equiv_functions_data)
                functions_data.append(orbit_functions_data)
            return functions_data

        # Add prim
        data["prim"] = self._prim.to_dict()

        # Add phenomenal
        data["phenomenal"] = (
            {"sites": self._phenomenal.to_list()}
            if self._phenomenal is not None
            else None
        )

        # Add generating_group (as indices)
        data["generating_group"] = self._generating_group.head_group_index

        # Add misc parameters...
        data.update(
            {
                "global_max_poly_order": self._global_max_poly_order,
                "orbit_branch_max_poly_order": self._orbit_branch_max_poly_order,
                "occ_site_basis_functions_specs": self._occ_site_basis_functions_specs,
                "make_equivalents": self._make_equivalents,
                "make_all_local_basis_sets": self._make_all_local_basis_sets,
                "make_variable_name_f": self._make_variable_name_f.to_dict(),
            }
        )

        # Add dof info...
        data.update(
            {
                "global_dof": self.global_dof,
                "local_continuous_dof": self.local_continuous_dof,
                "local_discrete_dof": self.local_discrete_dof,
                "local_dof": self.local_dof,
                "only_discrete": self.only_discrete,
            }
        )

        # Add site functions
        data.update(
            {
                "occ_site_functions": self.occ_site_functions,
                "occ_site_functions_info": self.occ_site_functions_info,
            }
        )

        # Add orbit_matrix_rep_builders
        data["orbit_matrix_rep_builders"] = [
            orbit_matrix_rep_builder.to_dict()
            for orbit_matrix_rep_builder in self.orbit_matrix_rep_builders
        ]

        # Add self.clusters
        data["clusters"] = _clusters_data(self.clusters)

        # Add constraints
        data["constraints"] = [
            [constraint.to_dict() for constraint in orbit_constraints]
            for orbit_constraints in self.constraints
        ]

        # Add self.functions
        data["functions"] = _functions_data(self.functions)

        # Add self.n_functions
        data["n_functions"] = self.n_functions

        # Add self.equivalent_clusters
        data["equivalent_clusters"] = (
            [
                _clusters_data(equiv_clusters)
                for equiv_clusters in self.equivalent_clusters
            ]
            if self.equivalent_clusters is not None
            else None
        )

        # Add self.equivalent_functions
        data["equivalent_functions"] = (
            [
                _functions_data(equiv_functions)
                for equiv_functions in self.equivalent_functions
            ]
            if self.equivalent_functions is not None
            else None
        )

        return data

    def _min_poly_order(self, orbit_prototype):
        if self.only_discrete:
            return len(orbit_prototype)
        return 1

    def _max_poly_order(self, orbit_prototype):
        if self.only_discrete:
            return len(orbit_prototype)
        branch = len(orbit_prototype)
        if branch in self._orbit_branch_max_poly_order:
            return self._orbit_branch_max_poly_order[branch]
        elif self._global_max_poly_order is not None:
            return int(self._global_max_poly_order)
        else:
            return branch

    def _build_prototype_basis_set(
        self,
        i_orbit: int,
        cluster: casmclust.Cluster,
    ):
        """Build the prototype cluster basis set

        Parameters
        ---------
        i_orbit: int
            Orbit number, starting from 0, used for printing information when
            `verbose` is set to `True`.
        cluster: libcasm.clusterography.Cluster
            A cluster in the orbit on which to build the cluster functions.

        Returns
        -------
        (prototype_basis_set, orbit_matrix_rep_builder, constraints):
            prototype_basis_set: list[PolynomialFunction]
                Symmetry adapted polynomial cluster functions for the prototype
                cluster in the orbit generated from `cluster`.

            orbit_matrix_rep_builder: casm.bset.cluster_functions.OrbitMatrixRepBuilder
                The OrbitMatrixRepBuilder contains matrix representations for generating
                polynomial cluster functions on the orbit prototype, coupling all local
                and global DoFs.

                Also includes:

                - Matrix representations for generating equivalent functions on other
                  clusters in the orbit.
                - Matrix representations for constructing local cluster functions for
                  symmetrically equivalent phenomenal clusters (local cluster functions
                  only).

            constraints: list[ExponentSumConstraint]
                The constraints used when constructing `prototype_basis_set` to ensure
                one and only one of mutually exclusive discrete variables
                is included in any monomial.
        """

        if self._verbose:
            print(f"### i_orbit: {i_orbit} ###")
            print()

            if self._phenomenal is not None:
                print("Phenomenal cluster:")
                print(xtal.pretty_json(self._phenomenal.to_dict(self._prim.xtal_prim)))
                print()

            print("Initial cluster:")
            print(xtal.pretty_json(cluster.to_dict(self._prim.xtal_prim)))
            print()

        builder = OrbitMatrixRepBuilder(
            prim=self._prim,
            generating_group=self._generating_group,
            global_dof=self.global_dof,
            local_continuous_dof=self.local_continuous_dof,
            local_discrete_dof=self.local_discrete_dof,
            cluster=cluster,
            phenomenal=self._phenomenal,
            make_variable_name_f=self._make_variable_name_f,
            occ_site_functions=self.occ_site_functions,
        )
        constraints = make_constraints(
            prototype_cluster=cluster,
            prototype_variables=builder.prototype_variables,
            prototype_variable_subsets=builder.prototype_variable_subsets,
            occ_site_functions=self.occ_site_functions,
            local_discrete_dof=self.local_discrete_dof,
        )

        if len(builder.prototype_variables) == 0:
            # if no variables, no functions
            if self._verbose:
                print("No variables")
            prototype_basis_set = []

        elif builder.n_local_variables == 0 and cluster.size() != 0:
            # if no local variables and not the null cluster, no functions
            # (only generate pure global DoF functions on the null cluster)
            if self._verbose:
                print("No local variables")
            prototype_basis_set = []

        else:
            if self._verbose:
                print("Variables:")
                for i_var, var in enumerate(builder.prototype_variables):
                    print(f"{i_var}: {var.name}")
                print()
                print("Variable subsets:")
                for i_subset, subset in enumerate(builder.prototype_variable_subsets):
                    print(f"{i_subset}:", end="")
                    for i_var in subset:
                        print(f"{builder.prototype_variables[i_var].name}, ", end="")
                    print()
                print()

            prototype_basis_set = make_symmetry_adapted_polynomials(
                matrix_rep=builder.prototype_matrix_rep,
                variables=builder.prototype_variables,
                variable_subsets=builder.prototype_variable_subsets,
                min_poly_order=self._min_poly_order(cluster),
                max_poly_order=self._max_poly_order(cluster),
                constraints=constraints,
                orthonormalize_in_place=False,
                verbose=self._verbose,
            )

            if self._verbose:
                print("Prototype cluster basis set:")
                if len(prototype_basis_set) == 0:
                    print("Empty")
                for i_func, func in enumerate(prototype_basis_set):
                    print(f"~~~ order: {func.order()}, function_index: {i_func} ~~~")
                    func._basic_print()
                    print()
                print()

        orbit_matrix_rep_builder = builder
        return (prototype_basis_set, orbit_matrix_rep_builder, constraints)

    def _build_orbit_basis_sets(
        self,
        i_orbit: int,
        orbit_matrix_rep_builder: OrbitMatrixRepBuilder,
        prototype_basis_set: list[PolynomialFunction],
    ):
        """Build basis sets for each cluster in an orbit

        Parameters
        ---------
        i_orbit: int
            Orbit number, starting from 0, used for printing information when
            `verbose` is set to `True`.
        orbit_matrix_rep_builder: casm.bset.cluster_functions.OrbitMatrixRepBuilder
            An OrbitMatrixRepBuilder which contains matrix representations of
            symmetry operations for generating symmetry-adapted polynomial cluster
            functions for the orbit from the orbit prototype, as output from
            :func:`~casm.bset.cluster_functions.ClusterFunctionsBuilder._build_prototype_basis_set`.
        prototype_basis_set: list[casm.bset.polynomial_functions.PolynomialFunction]
            Symmetry adapted polynomial cluster functions for the prototype
            cluster in the orbit, as output from
            :func:`~casm.bset.cluster_functions.ClusterFunctionsBuilder._build_prototype_basis_set`.

        Returns
        -------
        (orbit_basis_sets, orbit_clusters):

            orbit_basis_sets: list[list[\
            casm.bset.polynomial_functions.PolynomialFunction]]
                Symmetry adapted polynomial cluster functions for each cluster in
                the orbit, where ``orbit_basis_sets[i_equiv][i_func]``, is the
                `i_func`-th function on the `i_equiv`-th cluster in the orbit.

            orbit_clusters: list[libcasm.clusterography.Cluster]]:
                Orbit of clusters for which `orbit_basis_sets` has been constructed.

                The cluster ``orbit_clusters[i_equiv]`` is the `i_equiv`-th
                symmetrically equivalent cluster in the orbit.

                The order of sites in the clusters is not arbitrary, it is consistent
                with the `cluster_site_index` of the
                :class:`~casm.bset.polynomial_functions.Variable`
                used in the `orbit_basis_sets` functions.
        """
        builder = orbit_matrix_rep_builder

        orbit_basis_sets = []
        orbit_clusters = []

        # generate basis sets on all equivalent clusters
        if self._verbose:
            print("Generating equivalent cluster functions in orbit...")
            print()
        # i_equiv: index for equivalent clusters
        # M_list: sym rep matrices (1 for each sym op that prototype cluster to
        #     the equivalent cluster, only the first is needed here)
        for i_equiv, M_list in enumerate(builder.equivalence_map_inv_matrix_rep):
            # add equivalent cluster to orbit of clusters
            equiv_cluster = builder.equivalence_map_clusters[i_equiv][0]
            orbit_clusters.append(equiv_cluster)

            if self._verbose:
                print(f"~~~ i_orbit: {i_orbit}, i_equiv: {i_equiv} ~~~")
                print("Equivalent cluster:")
                print(xtal.pretty_json(equiv_cluster.to_dict(self._prim.xtal_prim)))
                print()

            # add basis set for equivalent cluster to orbit of basis sets
            equiv_basis_set = []
            for f_prototype in prototype_basis_set:
                M = M_list[0]
                S = FunctionRep(matrix_rep=M)
                f_equiv = S * f_prototype
                update_variables(
                    prim=self._prim,
                    cluster=equiv_cluster,
                    function=f_equiv,
                    prim_neighbor_list=self.prim_neighbor_list,
                    make_variable_name_f=self._make_variable_name_f,
                    local_discrete_dof=self.local_discrete_dof,
                )
                assert f_equiv.variables is not f_prototype.variables
                for i in range(len(f_prototype.variables)):
                    assert f_equiv.variables[i] is not f_prototype.variables[i]
                equiv_basis_set.append(f_equiv)
            orbit_basis_sets.append(equiv_basis_set)

            if self._verbose:
                print("Equivalent cluster basis set:")
                if len(equiv_basis_set) == 0:
                    print("Empty")
                for i_func, func in enumerate(equiv_basis_set):
                    print(f"~~~ order: {func.order()}, function_index: {i_func} ~~~")
                    func._basic_print()
                    print()
                print()

        return (orbit_basis_sets, orbit_clusters)

    def _build_equivalent_orbit_basis_sets(
        self,
        i_orbit: int,
        orbit_matrix_rep_builder: OrbitMatrixRepBuilder,
        orbit_basis_sets: list[list[PolynomialFunction]],
        orbit_clusters: list[casmclust.Cluster],
    ):
        """Build basis sets for each cluster in the equivalent orbits around \
        equivalent phenomenal cluster.

        Parameters
        ----------
        i_orbit: int
            Orbit number, starting from 0, used for printing information when
            `verbose` is set to `True`.

        orbit_matrix_rep_builder: casm.bset.cluster_functions.OrbitMatrixRepBuilder
            An OrbitMatrixRepBuilder which contains matrix representations of
            symmetry operations for generating polynomial cluster functions for the
            orbit from the orbit prototype, as output from
            :func:`~casm.bset.cluster_functions.ClusterFunctionsBuilder._build_prototype_basis_set`.

        orbit_basis_sets: list[list[casm.bset.polynomial_functions.PolynomialFunction]]
            Symmetry adapted polynomial cluster functions for each cluster in
            the orbit, where ``orbit_basis_sets[i_equiv][i_func]``, is the
            `i_func`-th function on the `i_equiv`-th cluster in the orbit.

        orbit_clusters: list[libcasm.clusterography.Cluster]]:
            Orbit of clusters for which `orbit_basis_sets` has been constructed.

            The cluster ``orbit_clusters[i_equiv]`` is the `i_equiv`-th
            symmetrically equivalent cluster in the orbit.

            The order of sites in the clusters is not arbitrary, it is consistent
            with the `cluster_site_index` of the
            :class:`~casm.bset.polynomial_functions.Variable`
            used in the `orbit_basis_sets` functions.

        Returns
        -------
        (equivalent_orbit_basis_sets, equivalent_orbit_clusters):

            equivalent_orbit_basis_sets: list[list[list[\
            casm.bset.polynomial_functions.PolynomialFunction]]]
                Symmetry adapted polynomial cluster functions for each cluster in each
                equivalent orbit about a symmetrically equivalent phenomenal cluster,
                where ``equivalent_orbit_basis_sets[i_clex][i_equiv][i_func]``, is the
                `i_func`-th function on the `i_equiv`-th cluster in the orbit about
                the `i_clex`-th equivalent phenomenal cluster.

            equivalent_orbit_clusters: list[list[libcasm.clusterography.Cluster]]]:
                Orbit of clusters for which `orbit_basis_sets` has been constructed.

                The cluster ``equivalent_orbit_clusters[i_clex][i_equiv]`` is the
                `i_equiv`-th symmetrically equivalent cluster in the orbit about the
                `i_clex`-th equivalent phenomenal cluster.

                The order of sites in the clusters is not arbitrary, it is consistent
                with the `cluster_site_index` of the
                :class:`~casm.bset.polynomial_functions.Variable`
                used in the `orbit_basis_sets` functions.

        """
        builder = orbit_matrix_rep_builder

        equivalent_orbit_basis_sets = []
        equivalent_orbit_clusters = []

        # generate equivalent local basis sets
        if self._verbose:
            print(
                "Generating local cluster functions "
                "for equivalent phenomenal clusters..."
            )
            print()

        # i_clex: index for equivalent local cluster expansion
        for i_clex in range(len(builder.phenomenal_generating_ops)):
            if self._verbose:
                print(f"*** i_orbit: {i_orbit}, i_clex: {i_clex} ***")
                print("Equivalent phenomenal cluster:")
                site_rep = builder.phenomenal_generating_site_rep[i_clex]
                equiv_phenomenal = site_rep * self.phenomenal
                print(xtal.pretty_json(equiv_phenomenal.to_dict(self._prim.xtal_prim)))
                print()

            _equiv_orbit_basis_sets = []
            _equiv_orbit_clusters = []
            # i_clust: index for equivalent clusters
            # M: sym rep matrix for transforming functions on the `i_clust`-th cluster
            #    to the orbit around the `i_clex`-th equivalent phenomenal cluster
            for i_clust, M in enumerate(
                builder.phenomenal_generating_inv_matrix_rep[i_clex]
            ):
                # add equivalent cluster to orbit of clusters
                site_rep = builder.phenomenal_generating_site_rep[i_clex]
                equiv_cluster = site_rep * orbit_clusters[i_clust]
                _equiv_orbit_clusters.append(equiv_cluster)

                if self._verbose:
                    print(
                        f"~~~ i_orbit: {i_orbit}, "
                        f"i_clex: {i_clex}, "
                        f"i_clust: {i_clust} ~~~"
                    )
                    print("Equivalent cluster:")
                    print(xtal.pretty_json(equiv_cluster.to_dict(self._prim.xtal_prim)))
                    print()

                # add basis set for equivalent cluster to orbit of basis sets
                equiv_basis_set = []
                for f_prototype in orbit_basis_sets[i_clust]:
                    S = FunctionRep(matrix_rep=M)
                    f_equiv = S * f_prototype
                    update_variables(
                        prim=self._prim,
                        cluster=equiv_cluster,
                        function=f_equiv,
                        prim_neighbor_list=self.prim_neighbor_list,
                        make_variable_name_f=self._make_variable_name_f,
                        local_discrete_dof=self.local_discrete_dof,
                    )
                    assert f_equiv.variables is not f_prototype.variables
                    for i in range(len(f_prototype.variables)):
                        assert f_equiv.variables[i] is not f_prototype.variables[i]
                    equiv_basis_set.append(f_equiv)
                _equiv_orbit_basis_sets.append(equiv_basis_set)

                if self._verbose:
                    print("Equivalent cluster basis set:")
                    if len(equiv_basis_set) == 0:
                        print("Empty")
                    for i_func, func in enumerate(equiv_basis_set):
                        print(
                            f"~~~ order: {func.order()}, function_index: {i_func} ~~~"
                        )
                        func._basic_print()
                        print()
                    print()

            equivalent_orbit_basis_sets.append(_equiv_orbit_basis_sets)
            equivalent_orbit_clusters.append(_equiv_orbit_clusters)

        return (equivalent_orbit_basis_sets, equivalent_orbit_clusters)

    def basis_dict(
        self,
        clex_basis_specs: ClexBasisSpecs,
        coordinate_mode: str = "frac",
    ) -> dict:
        R"""Generate the CASM basis.json data

        Parameters
        ----------
        clex_basis_specs: casm.bset.cluster_functions.ClexBasisSpecs
            The specifications for the cluster expansion basis set.
        coordinate_mode: str = "frac"
            The coordinate mode used to represent cluster invariant group operations.
            The default value is "frac". The other option is "cart".

        Returns
        -------
        data: dict
            A description of the generated cluster expansion basis set, as described
            `here <https://prisms-center.github.io/CASMcode_docs/formats/casm/clex/ClexBasis/>`_.
        """  # noqa
        prim = self._prim
        clusters = self.clusters
        functions = self.functions
        occ_site_functions = self.occ_site_functions

        basis_data = {}
        basis_data["bspecs"] = clex_basis_specs.to_dict()
        basis_data["prim"] = prim.xtal_prim.to_dict()

        basis_data["orbits"] = _make_orbits_data(
            prim=self._prim,
            clex_basis_specs=clex_basis_specs,
            clusters=clusters,
            functions=functions,
            coordinate_mode=coordinate_mode,
        )

        basis_data["site_functions"] = _make_site_functions_data(
            prim=prim,
            occ_site_functions=occ_site_functions,
        )

        return basis_data

    def equivalents_info_dict(self) -> dict:
        R"""Generate the CASM equivalents_info.json data

        Returns
        -------
        equivalents_info_dict: dict
            The equivalents info provides the phenomenal cluster and local-cluster
            orbits for all symmetrically equivalent local-cluster expansions, and the
            indices of the factor group operations used to construct each equivalent
            local cluster expansion from the prototype local-cluster expansion. When
            there is an orientation to the local-cluster expansion this information
            allows generating the proper diffusion events, etc. from the prototype.

            A description of the generated cluster expansion basis set, as described
            `here <TODO>`_.
        """  # noqa

        # Equivalents info, prototype
        equivalents_info = {}

        if self._phenomenal is None:
            return equivalents_info
        if len(self.orbit_matrix_rep_builders) == 0:
            raise Exception(
                "Error in ClusterFunctionsBuilder.equivalents_info_dict: No orbits"
            )

        # Write prim factor group info
        equivalents_info[
            "factor_group"
        ] = config_io.symgroup_to_dict_with_group_classification(
            self._prim, self._prim.factor_group
        )

        # Write equivalents generating ops
        # (actually prim factor group indices of those ops and the
        #  translations can be figured out from the phenomenal cluster)
        orbit_matrix_rep_builder = self.orbit_matrix_rep_builders[0]
        equivalents_info[
            "equivalent_generating_ops"
        ] = orbit_matrix_rep_builder.phenomenal_generating_indices

        # Write prototype orbits info
        tmp = {}
        prototype_phenomenal = self._phenomenal
        clusters = self.clusters
        tmp["phenomenal"] = prototype_phenomenal.to_dict(xtal_prim=self._prim.xtal_prim)
        tmp["prim"] = self._prim.to_dict()
        tmp["orbits"] = orbits_to_dict(
            orbits=clusters,
            prim=self._prim,
        )
        equivalents_info["prototype"] = tmp

        # Write equivalent orbits info
        equivalents_info["equivalents"] = []
        for i_clex, clusters in enumerate(self.equivalent_clusters):
            tmp = {}
            site_rep = orbit_matrix_rep_builder.phenomenal_generating_site_rep[i_clex]
            equiv_phenomenal = site_rep * prototype_phenomenal
            tmp["phenomenal"] = equiv_phenomenal.to_dict(xtal_prim=self._prim.xtal_prim)
            tmp["prim"] = self._prim.to_dict()
            tmp["orbits"] = orbits_to_dict(
                orbits=clusters,
                prim=self._prim,
            )
            equivalents_info["equivalents"].append(tmp)

        return equivalents_info


def make_point_functions(
    prim: casmconfig.Prim,
    prim_neighbor_list: PrimNeighborList,
    orbit: list[casmclust.Cluster],
    orbit_functions: list[list[PolynomialFunction]],
    make_variable_name_f: Optional[Callable] = None,
    local_discrete_dof: Optional[dict] = None,
):
    """Construct point functions

    This method uses the data generated by :class:`ClusterFunctionsBuilder` to
    collect all the functions that involve each point. These functions are used
    by Clexulator methods that can calculate point correlations and changes in point
    correlations.

    Parameters
    ----------
    prim_neighbor_list: libcasm.clexulator.PrimNeighborList
        The :class:`PrimNeighborList` is used to uniquely index sites with local
        variables included in the cluster functions, relative to a reference unit cell.

    orbit: list[libcasm.clusterography.Cluster]
        An orbit of clusters, where ``orbit[i_equiv]``, is the
        `i_equiv`-th symmetrically equivalent cluster in the orbit.
        The order of sites in the clusters may not be arbitrary, it must be
        consistent with the `cluster_site_index` of the :class:`Variable` used
        in the :class:`PolynomialFunction` returned in `orbit_functions`.

        This should generally be one orbit as generated by
        :class:`ClusterFunctionsBuilder`.

    orbit_functions: list[list[casm.bset.polynomial_functions.PolynomialFunction]]
        Polynomial functions, where ``functions[i_equiv][i_func]``,
        is the `i_func`-th function on the cluster given by `orbit[i_equiv]`.

        This should generally be the functions for one orbit as generated by
        :class:`ClusterFunctionsBuilder`.

    make_variable_name_f: Optional[Callable] = None
        Allows specifying a custom class to construct variable names. The default
        class used is :class:`~casm.bset.cluster_functions.MakeVariableName`.
        Custom classes should have the same `__call__` signature as
        :class:`~casm.bset.cluster_functions.MakeVariableName`, and have
        `occ_var_name` and `occ_var_desc` attributes.

    local_discrete_dof: Optionla[list[str]] = None
        The types of local discrete degree of freedom (DoF).

    Returns
    -------
    point_functions: list[list[list[casm.bset.polynomial_functions.PolynomialFunction]]]
        Polynomial functions, where ``point_functions[i_func][nlist_index]`` is
        a list of PolynomialFunction that are symmetrically equivalent to the
        `i_func`-th function on the clusters and involve the `nlist_index`-th site
        in the neighbor list.

    """
    if len(orbit) != len(orbit_functions):
        raise Exception(
            "Error in make_point_functions: "
            "orbit size does not match orbit_functions size"
        )

    if not len(orbit):
        return [[[]]]

    points = []
    for sublattice_index in prim_neighbor_list.sublattice_indices():
        points.append(
            xtal.IntegralSiteCoordinate(
                sublattice=sublattice_index,
                unitcell=[0, 0, 0],
            )
        )

    # orbit_point_functions: list[list[list[PolynomialFunction]]]
    # orbit_point_functions[i_func][nlist_index][i_point_function]
    point_functions = []
    for i_func in range(len(orbit_functions[0])):
        equiv_functions = []
        for nlist_index, point in enumerate(points):
            equiv_functions_by_point = []
            for i_equiv, equiv in enumerate(orbit):
                # for each cluster site on same sublattice as point,
                for site in equiv:
                    if site.sublattice() != point.sublattice():
                        continue
                    f = orbit_functions[i_equiv][i_func].copy()
                    update_variables(
                        prim=prim,
                        cluster=equiv,
                        function=f,
                        prim_neighbor_list=prim_neighbor_list,
                        translation=-site.unitcell(),
                        make_variable_name_f=make_variable_name_f,
                        local_discrete_dof=local_discrete_dof,
                    )
                    # add to point functions
                    equiv_functions_by_point.append(f)
            equiv_functions.append(equiv_functions_by_point)
        point_functions.append(equiv_functions)

    return point_functions


def make_local_point_functions(
    prim_neighbor_list: PrimNeighborList,
    orbit: list[casmclust.Cluster],
    orbit_functions: list[list[PolynomialFunction]],
):
    """Construct local cluster point functions

    This method uses the data generated by :class:`ClusterFunctionsBuilder` to
    collect all the local cluster functions that involve each point. These functions
    are used by Clexulator methods that can calculate point correlations and changes
    in point correlations.

    Parameters
    ----------
    prim_neighbor_list: libcasm.clexulator.PrimNeighborList
        The :class:`PrimNeighborList` is used to uniquely index sites with local
        variables included in the cluster functions, relative to a reference unit cell.

    orbit: list[libcasm.clusterography.Cluster]
        An orbit of clusters, where ``orbit[i_equiv]``, is the
        `i_equiv`-th symmetrically equivalent cluster in the orbit.
        The order of sites in the clusters may not be arbitrary, it must be
        consistent with the `cluster_site_index` of the :class:`Variable` used
        in the :class:`PolynomialFunction` returned in `orbit_functions`.

        This should generally be one orbit as generated by
        :class:`ClusterFunctionsBuilder`.

    orbit_functions: list[list[casm.bset.polynomial_functions.PolynomialFunction]]
        Polynomial functions, where ``functions[i_equiv][i_func]``,
        is the `i_func`-th function on the cluster given by `orbit[i_equiv]`.

        This should generally be the functions for one orbit as generated by
        :class:`ClusterFunctionsBuilder`.

    Returns
    -------
    point_functions: list[list[list[casm.bset.polynomial_functions.PolynomialFunction]]]
        Polynomial functions, where ``point_functions[i_func][nlist_index]`` is
        a list of PolynomialFunction that are symmetrically equivalent to the
        `i_func`-th function on the clusters and involve the `nlist_index`-th site
        in the neighbor list.

    """
    if len(orbit) != len(orbit_functions):
        raise Exception(
            "Error in make_local_point_functions: "
            "orbit size does not match orbit_functions size"
        )

    if not len(orbit):
        return [[[]]]

    orbit_neighbor_indices = set()
    for equiv in orbit:
        for site in equiv:
            nlist_index = prim_neighbor_list.neighbor_index(site)
            orbit_neighbor_indices.add(nlist_index)

    # orbit_point_functions: list[list[list[PolynomialFunction]]]
    # orbit_point_functions[i_func][nlist_index][i_point_function]
    prototype_basis_set = orbit_functions[0]
    point_functions = []
    for i_func in range(len(prototype_basis_set)):
        point_functions.append([])
        for nlist_index in range(max(orbit_neighbor_indices) + 1):
            point_functions[i_func].append([])

    for i_equiv, equiv_functions in enumerate(orbit_functions):
        for site in orbit[i_equiv]:
            nlist_index = prim_neighbor_list.neighbor_index(site)
            for i_func, f in enumerate(equiv_functions):
                point_functions[i_func][nlist_index].append(f)

    return point_functions



import copy


def almost_zero(value, abs_tol=1e-5) -> bool:
    """Check if value is approximately zero, using an absolute tolerance"""
    return abs(value) < abs_tol


def almost_equal(value1, value2, abs_tol=1e-5) -> bool:
    """Check if two values are approximately equal, using an absolute tolerance"""
    return almost_zero(value1 - value2, abs_tol=abs_tol)


def almost_int(value, abs_tol=1e-5) -> bool:
    """Check if a floating point value is approximately integer, using an \
    absolute tolerance"""
    return almost_zero(abs(value - round(value)), abs_tol=abs_tol)


def signof(value, abs_tol=1e-5) -> float:
    """Return the sign of a floating point value (or zero)

    Parameters
    ----------
    value: float
        The value to check.
    abs_tol: float = 1e-5
        The absolute tolerance to check for zero.

    Returns
    -------
    signof: float
        One of -1.0, 0.0, or 1.0.

    """
    if value < -abs_tol:
        return -1.0
    elif value > abs_tol:
        return 1.0
    else:
        return 0.0


def irrational_to_tex_string(
    value: float,
    limit: int,
    max_pow: int,
    abs_tol: 1e-5,
) -> str:
    """Find irrational approximation and return as a tex string

    Finds best irrational number approximation of a floating point value and
    returns a tex-formated string that contains an irrational approximation.

    This method searches numbers of the form (x/y)^(1/z), where x and y range
    from 1 to `limit` and z ranges from 1 to `max_pow`.

    Notes
    -----
    This is a pure Python reproduction of the C++ function irrational_to_tex_string
    found in libcasm-global.

    Parameters
    ----------
    value: float
        A floating point number.
    limit: int
        Searches for `x` and `y` in the range `[1, limit]`.
    max_pow: int
        Search for `z` in range `[1, max_pow]`.

    Returns
    -------
    texstring: str
        The approximation, :math:`(x/y)^(1/z)` of `value`, as a tex-formatted string.
    """

    texstring = ""
    if almost_int(value, abs_tol=abs_tol):
        return str(round(value))
    if value < 0.0:
        texstring += "-"
        value = abs(value)

    tmp_value = copy.copy(value)
    tmp_denom = 0
    tmp_num = 0
    y = 0
    x = 0

    z = 1
    while z < max_pow + 1:
        i = 1
        while i < limit + 1:
            tmp_denom = i / tmp_value
            tmp_num = tmp_value / i
            if tmp_denom > 1.0 and almost_int(tmp_denom, abs_tol=abs_tol):
                x = i
                y = round(tmp_denom)
            elif tmp_num > 1.0 and almost_int(tmp_num, abs_tol=abs_tol):
                y = i
                x = round(tmp_num)
            else:
                i += 1
                continue

            if z == 1:
                texstring += str(x) + "/" + str(y)
                return texstring

            if z == 2:
                texstring += "\\sqrt{" + str(x)
                if y != 1:
                    texstring += "/" + str(y)
                texstring += "}"
                return texstring
            else:
                texstring += "(" + str(x)
                if y != 1:
                    texstring += "/" + str(y)
                texstring += ")^{1/" + str(z) + "}"
                return texstring

            i += 1
        tmp_value *= value
        z += 1
    texstring += str(value)
    return texstring


def factor_by_mode(x, abs_tol=1e-5):
    """Factor array of float by the mode value

    Parameters
    ----------
    x: np.ndarray[np.float]
        Array of float
    abs_tol: float
        Absolute tolerance for comparison

    Returns
    -------
    mode: float
        The most commonly occurring value in `x`.
    x_factored: np.ndarray[np.float]
        The input array, `x`, divided by the mode.
    """
    counts = []
    for _x in x:
        found = False
        for i in range(len(counts)):
            key, count = counts[i]
            if almost_equal(_x, key, abs_tol=abs_tol):
                counts[i] = (key, count + 1)
                found = True
                break
        if not found:
            counts.append((_x, 1))
    max_count = None
    mode = None
    for i in range(len(counts)):
        key, count = counts[i]
        if max_count is None:
            max_count = count
            mode = key
        else:
            if count > max_count:
                max_count = count
                mode = key
    return (mode, x / mode)



from typing import TypeVar

import libcasm.configuration as casmconfig
from casm.bset.parsing import (
    required_from_dict,
)
from libcasm.clusterography import ClusterSpecs

from ._basis_function_specs import (
    BasisFunctionSpecs,
)

ClexBasisSpecsType = TypeVar("ClexBasisSpecs")


class ClexBasisSpecs:
    """Cluster expansion basis set specifications

    A cluster expansion basis set is specified by two components, one specifying
    basis function type and order, and one specifying cluster orbits.

    The `ClexBasisSpecs format`_ is used for the `bspecs.json` file in a CASM project.

    .. _`ClexBasisSpecs format`: https://prisms-center.github.io/CASMcode_docs/formats/casm/clex/ClexBasisSpecs/

    """

    def __init__(
        self,
        cluster_specs: ClusterSpecs,
        basis_function_specs: BasisFunctionSpecs,
    ):
        """
        .. rubric:: Constructor

        Parameters
        ----------
        cluster_specs: libcasm.clusterography.ClusterSpecs
            Specifies which cluster orbits will be included in the cluster expansion
            basis functions.
        basis_function_specs: casm.bset.cluster_functions.BasisFunctionSpecs
            Specifies how the cluster expansion basis function type and order.
        """
        self.cluster_specs = cluster_specs
        """libcasm.clusterography.ClusterSpecs: Cluster orbit specifications
        
        Specifies which cluster orbits will be included in the cluster expansion 
        basis functions.
        """

        self.basis_function_specs = basis_function_specs
        """casm.bset.cluster_functions.BasisFunctionSpecs: Basis function specifications
        
        Specifies the cluster expansion basis function type and order.
        """

    def is_periodic(self):
        if self.cluster_specs.phenomenal() is None:
            return True
        return False

    @staticmethod
    def from_dict(
        data: dict,
        prim: casmconfig.Prim,
    ) -> ClexBasisSpecsType:
        """Construct ClexBasisSpecs from a Python dictionary

        See the CASM documentation for the `ClexBasisSpecs format`_.

        .. _`ClexBasisSpecs format`: https://prisms-center.github.io/CASMcode_docs/formats/casm/clex/ClexBasisSpecs/


        Parameters
        ----------
        data : dict
            A Python dictionary formatted as documented
        prim: libcasm.configuration.Prim
            The prim, with symmetry information.

        Returns
        -------
        clex_basis_specs : ClexBasisSpecs
            The created ClexBasisSpecs
        """
        if "cluster_specs" not in data:
            raise Exception("Error in ClexBasisSpecs.from_dict: no cluster_specs")
        if "params" in data["cluster_specs"]:
            # v1 format:
            cluster_specs_data = data["cluster_specs"]
            cluster_specs_key = "params"
        else:
            # v2 format:
            cluster_specs_data = data
            cluster_specs_key = "cluster_specs"

        return ClexBasisSpecs(
            cluster_specs=required_from_dict(
                ClusterSpecs,
                cluster_specs_data,
                cluster_specs_key,
                xtal_prim=prim.xtal_prim,
                prim_factor_group=prim.factor_group,
                integral_site_coordinate_symgroup_rep=prim.integral_site_coordinate_symgroup_rep,
            ),
            basis_function_specs=required_from_dict(
                BasisFunctionSpecs, data, "basis_function_specs"
            ),
        )

    def to_dict(self) -> dict:
        """Represent the ClexBasisSpecs as a Python dictionary

        See the CASM documentation for the `ClexBasisSpecs format`_.

        .. _`ClexBasisSpecs format`: https://prisms-center.github.io/CASMcode_docs/formats/casm/clex/ClexBasisSpecs/


        Returns
        -------
        data : dict
            The ClexBasisSpecs as a Python dictionary

        """
        return {
            "cluster_specs": self.cluster_specs.to_dict(),
            "basis_function_specs": self.basis_function_specs.to_dict(),
        }
